"""`sample.py`.

This is a sample python script showing how to train an end-to-end OTX Anomaly Classification Task.
"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import importlib
import os
import shutil
from argparse import Namespace
from typing import Any, Dict, Optional, Type, Union

from otx.algorithms.anomaly.adapters.anomalib.data.dataset import (
    AnomalyClassificationDataset,
    AnomalyDetectionDataset,
    AnomalySegmentationDataset,
)
from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.algorithms.anomaly.tasks import NNCFTask, OpenVINOTask
from otx.api.configuration.helper import create as create_hyper_parameters
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType, parse_model_template
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.train_parameters import TrainParameters
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType

logger = get_logger(__name__)


# pylint: disable=too-many-instance-attributes
class OtxAnomalyTask:
    """OTX Anomaly Classification Task."""

    def __init__(
        self,
        dataset_path: str,
        train_subset: Dict[str, str],
        val_subset: Dict[str, str],
        test_subset: Dict[str, str],
        model_template_path: str,
        seed: Optional[int] = None,
    ) -> None:
        """Initialize OtxAnomalyTask.

        Args:
            dataset_path (str): Path to the MVTec dataset.
            train_subset (Dict[str, str]): Dictionary containing path to train annotation file and path to dataset.
            val_subset (Dict[str, str]): Dictionary containing path to validation annotation file and path to dataset.
            test_subset (Dict[str, str]): Dictionary containing path to test annotation file and path to dataset.
            model_template_path (str): Path to model template.
            seed (Optional[int]): Setting seed to a value other than 0 also marks PytorchLightning trainer's
                deterministic flag to True.

        Example:
            >>> import os
            >>> os.getcwd()
            '~/otx/external/anomaly'

            If MVTec dataset is placed under the above directory, then we could run,

            >>> model_template_path = "./configs/classification/padim/template.yaml"
            >>> dataset_path = "./datasets/MVTec"
            >>> task = OtxAnomalyTask(
            ...     dataset_path=dataset_path,
            ...     train_subset={"ann_file": train.json, "data_root": dataset_path},
            ...     val_subset={"ann_file": val.json, "data_root": dataset_path},
            ...     test_subset={"ann_file": test.json, "data_root": dataset_path},
            ...     model_template_path=model_template_path
            ... )

            >>> task.train()
            Performance(score: 1.0, dashboard: (1 metric groups))

            >>> task.export()
            Performance(score: 0.9756097560975608, dashboard: (1 metric groups))
        """
        logger.info("Loading the model template.")
        self.model_template = parse_model_template(model_template_path)

        logger.info("Loading MVTec dataset.")
        self.task_type = self.model_template.task_type

        dataclass = self.get_dataclass()

        self.dataset = dataclass(train_subset, val_subset, test_subset)

        logger.info("Creating the task-environment.")
        self.task_environment = self.create_task_environment()

        logger.info("Creating the base Torch and OpenVINO tasks.")
        self.torch_task = self.create_task(task="base")

        self.trained_model: ModelEntity
        self.openvino_task: OpenVINOTask
        self.nncf_task: NNCFTask
        self.results = {"category": dataset_path}
        self.seed = seed

    def get_dataclass(
        self,
    ) -> Union[Type[AnomalyDetectionDataset], Type[AnomalySegmentationDataset], Type[AnomalyClassificationDataset]]:
        """Gets the dataloader based on the task type.

        Raises:
            ValueError: Validates task type.

        Returns:
           Dataloader
        """
        dataclass: Union[
            Type[AnomalyDetectionDataset], Type[AnomalySegmentationDataset], Type[AnomalyClassificationDataset]
        ]
        if self.task_type == TaskType.ANOMALY_DETECTION:
            dataclass = AnomalyDetectionDataset
        elif self.task_type == TaskType.ANOMALY_SEGMENTATION:
            dataclass = AnomalySegmentationDataset
        elif self.task_type == TaskType.ANOMALY_CLASSIFICATION:
            dataclass = AnomalyClassificationDataset
        else:
            raise ValueError(f"{self.task_type} not a supported task")
        return dataclass

    def create_task_environment(self) -> TaskEnvironment:
        """Create task environment."""
        hyper_parameters = create_hyper_parameters(self.model_template.hyper_parameters.data)
        labels = self.dataset.get_labels()
        label_schema = LabelSchemaEntity.from_labels(labels)

        return TaskEnvironment(
            model_template=self.model_template,
            model=None,
            hyper_parameters=hyper_parameters,
            label_schema=label_schema,
        )

    def create_task(self, task: str) -> Any:
        """Create base torch or openvino task.

        Args:
            task (str): task type. Either base or openvino.

        Returns:
            Any: Base Torch or OpenVINO Task Class.

        Example:
            >>> self.create_task(task="base")
            <anomaly_classification.torch_task.AnomalyClassificationTask>

        """
        if self.model_template.entrypoints is not None:
            task_path = getattr(self.model_template.entrypoints, task)
        else:
            raise ValueError(f"Cannot create {task} task. `model_template.entrypoint` does not have {task}")

        module_name, class_name = task_path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        return getattr(module, class_name)(task_environment=self.task_environment)

    def train(self) -> ModelEntity:
        """Train the base Torch model."""
        logger.info("Training the model.")
        output_model = ModelEntity(
            train_dataset=self.dataset,
            configuration=self.task_environment.get_model_configuration(),
        )
        self.torch_task.train(
            dataset=self.dataset, output_model=output_model, train_parameters=TrainParameters(), seed=self.seed
        )

        logger.info("Inferring the base torch model on the validation set.")
        result_set = self.infer(self.torch_task, output_model)

        logger.info("Evaluating the base torch model on the validation set.")
        self.evaluate(self.torch_task, result_set)
        self.results["torch_fp32"] = result_set.performance.score.value
        self.trained_model = output_model
        return self.trained_model

    def infer(self, task: IInferenceTask, output_model: ModelEntity) -> ResultSetEntity:
        """Get the predictions using the base Torch or OpenVINO tasks and models.

        Args:
            task (IInferenceTask): Task to infer. Either torch or openvino.
            output_model (ModelEntity): Output model on which the weights are saved.

        Returns:
            ResultSetEntity: Results set containing the true and pred datasets.

        """
        ground_truth_validation_dataset = self.dataset.get_subset(Subset.VALIDATION)
        prediction_validation_dataset = task.infer(
            dataset=ground_truth_validation_dataset.with_empty_annotations(),
            inference_parameters=InferenceParameters(is_evaluation=True),
        )

        return ResultSetEntity(
            model=output_model,
            ground_truth_dataset=ground_truth_validation_dataset,
            prediction_dataset=prediction_validation_dataset,
        )

    @staticmethod
    def evaluate(task: IEvaluationTask, result_set: ResultSetEntity) -> None:
        """Evaluate the performance of the model.

        Args:
            task (IEvaluationTask): Task to evaluate the performance. Either torch or openvino.
            result_set (ResultSetEntity): Results set containing the true and pred datasets.

        """
        task.evaluate(result_set)
        logger.info(str(result_set.performance))

    def export(self) -> ModelEntity:
        """Export the model via openvino."""
        logger.info("Exporting the model.")
        exported_model = ModelEntity(
            train_dataset=self.dataset,
            configuration=self.task_environment.get_model_configuration(),
        )
        self.torch_task.export(ExportType.OPENVINO, exported_model)
        self.task_environment.model = exported_model

        logger.info("Creating the OpenVINO Task.")

        self.openvino_task = self.create_task(task="openvino")

        logger.info("Inferring the exported model on the validation set.")
        result_set = self.infer(task=self.openvino_task, output_model=exported_model)

        logger.info("Evaluating the exported model on the validation set.")
        self.evaluate(task=self.openvino_task, result_set=result_set)
        self.results["vino_fp32"] = result_set.performance.score.value

        return exported_model

    def optimize(self) -> None:
        """Optimize the model via POT."""
        logger.info("Running the POT optimization")
        optimized_model = ModelEntity(
            self.dataset,
            configuration=self.task_environment.get_model_configuration(),
        )

        self.openvino_task.optimize(
            optimization_type=OptimizationType.POT,
            dataset=self.dataset,
            output_model=optimized_model,
            optimization_parameters=OptimizationParameters(),
        )

        logger.info("Inferring the optimised model on the validation set.")
        result_set = self.infer(task=self.openvino_task, output_model=optimized_model)

        logger.info("Evaluating the optimized model on the validation set.")
        self.evaluate(task=self.openvino_task, result_set=result_set)
        self.results["pot_int8"] = result_set.performance.score.value

    def optimize_nncf(self) -> None:
        """Optimize the model via NNCF."""
        logger.info("Running the NNCF optimization")
        init_model = ModelEntity(
            self.dataset,
            configuration=self.task_environment.get_model_configuration(),
            model_adapters={"weights.pth": ModelAdapter(self.trained_model.get_data("weights.pth"))},
        )

        self.task_environment.model = init_model
        self.nncf_task = self.create_task("nncf")

        optimized_model = ModelEntity(
            self.dataset,
            configuration=self.task_environment.get_model_configuration(),
        )
        self.nncf_task.optimize(OptimizationType.NNCF, self.dataset, optimized_model)

        logger.info("Inferring the optimised model on the validation set.")
        result_set = self.infer(task=self.nncf_task, output_model=optimized_model)

        logger.info("Evaluating the optimized model on the validation set.")
        self.evaluate(task=self.nncf_task, result_set=result_set)
        self.results["torch_int8"] = result_set.performance.score.value

    def export_nncf(self) -> ModelEntity:
        """Export NNCF model via openvino."""
        logger.info("Exporting the model.")
        exported_model = ModelEntity(
            train_dataset=self.dataset,
            configuration=self.task_environment.get_model_configuration(),
        )
        self.nncf_task.export(ExportType.OPENVINO, exported_model)
        self.task_environment.model = exported_model

        logger.info("Creating the OpenVINO Task.")

        self.openvino_task = self.create_task(task="openvino")

        logger.info("Inferring the exported model on the validation set.")
        result_set = self.infer(task=self.openvino_task, output_model=exported_model)

        logger.info("Evaluating the exported model on the validation set.")
        self.evaluate(task=self.openvino_task, result_set=result_set)
        self.results["vino_int8"] = result_set.performance.score.value
        return exported_model

    @staticmethod
    def clean_up() -> None:
        """Clean up the `results` directory used by `anomalib`."""
        results_dir = "./results"
        if os.path.exists(results_dir):
            shutil.rmtree(results_dir)


def parse_args() -> Namespace:
    """Parse CLI arguments.

    Returns:
        (Namespace): CLI arguments.

    """
    parser = argparse.ArgumentParser(
        description="Sample showcasing how to run Anomaly Classification Task using OTX SDK"
    )
    parser.add_argument(
        "--model_template_path",
        default="./configs/classification/padim/template.yaml",
    )
    parser.add_argument("--dataset_path", default="./datasets/MVTec")
    parser.add_argument("--category", default="bottle")
    parser.add_argument("--train-ann-files", required=True)
    parser.add_argument("--val-ann-files", required=True)
    parser.add_argument("--test-ann-files", required=True)
    parser.add_argument("--optimization", choices=("none", "pot", "nncf"), default="none")
    parser.add_argument("--seed", default=0)
    return parser.parse_args()


def main() -> None:
    """Run `sample.py` with given CLI arguments."""
    args = parse_args()
    path = os.path.join(args.dataset_path, args.category)

    train_subset = {"ann_file": args.train_ann_files, "data_root": path}
    val_subset = {"ann_file": args.val_ann_files, "data_root": path}
    test_subset = {"ann_file": args.test_ann_files, "data_root": path}

    task = OtxAnomalyTask(
        dataset_path=path,
        train_subset=train_subset,
        val_subset=val_subset,
        test_subset=test_subset,
        model_template_path=args.model_template_path,
        seed=args.seed,
    )

    task.train()
    task.export()

    if args.optimization == "pot":
        task.optimize()

    if args.optimization == "nncf":
        task.optimize_nncf()
        task.export_nncf()

    task.clean_up()


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import torch
import torch.nn as nn
from mmcls.models.builder import LOSSES
from torch import Tensor


def off_diagonal(x: Tensor):
    """
    return a tensor containing all the elements outside the diagonal of x
    """
    assert x.shape[0] == x.shape[1]
    return x.flatten()[:-1].view(x.shape[0] - 1, x.shape[0] + 1)[:, 1:].flatten()


@LOSSES.register_module()
class BarlowTwinsLoss(nn.Module):
    """
    Barlow Twins Loss: https://arxiv.org/abs/2103.03230.
    Self-Supervised Learning via Redundancy Reduction
    Code adapted from https://github.com/facebookresearch/barlowtwins.
    """

    def __init__(self, off_diag_penality, loss_weight=1.0):
        super().__init__()
        self.penalty = off_diag_penality
        self.loss_weight = loss_weight

    def forward(self, feats1: Tensor, feats2: Tensor, **kwargs):
        """
        Compute Barlow Twins Loss and, if labels are not none,
        also the Cross-Entropy loss.
        Args:
            feats1, feats2: vectors of shape [bsz, ...]. Corresponding to
                            two views of the same samples
        Returns:
            A floating point number describing the Barlow Twins loss
        """

        batch_size = feats1.shape[0]
        assert batch_size == feats2.shape[0]
        dimensionality = feats1.shape[1]
        assert dimensionality == feats2.shape[1]

        # Barlow Twins loss: redundancy reduction
        batch_norm = nn.BatchNorm1d(dimensionality, affine=False, track_running_stats=False)
        # empirical cross-correlation matrix
        eccm = batch_norm(feats1).T @ batch_norm(feats2)
        eccm.div_(batch_size)

        # Compute the invariance term (diagonal) and redundacy term (off-diagonal)
        on_diag = torch.diagonal(eccm).add(-1).pow_(2).sum()
        off_diag = off_diagonal(eccm).pow_(2).sum()
        # Normalize the loss by the dimensionality of the projector
        return self.loss_weight * (on_diag + self.penalty * off_diag) / dimensionality



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from os import path as osp

import mmcv
import numpy as np
import torch
from mmcv.runner import HOOKS, EvalHook, Hook
from torch.utils.data import DataLoader


@HOOKS.register_module()
class CustomEvalHook(EvalHook):
    """Custom Evaluation hook for the MPA

    Args:
        dataloader (DataLoader): A PyTorch dataloader.
        interval (int): Evaluation interval (by epochs). Default: 1.
    """

    def __init__(
        self,
        *args,
        ema_eval_start_epoch=10,
        **kwargs,
    ):
        metric = kwargs["metric"]
        self.metric = None
        if isinstance(metric, str):
            self.metric = "top-1" if metric == "accuracy" else metric
        else:
            self.metric = metric[0]
            if metric.count("class_accuracy") > 0:
                self.metric = "accuracy"
            elif metric.count("accuracy") > 0:
                self.metric = "top-1"
        super().__init__(*args, **kwargs, save_best=self.metric, rule="greater")
        self.ema_eval_start_epoch = ema_eval_start_epoch

        self.best_loss = 9999999.0
        self.best_score = 0.0
        self.save_mode = self.eval_kwargs.get("save_mode", "score")

    def _do_evaluate(self, runner, ema=False):
        """perform evaluation"""
        results = single_gpu_test(runner.model, self.dataloader)
        if ema and hasattr(runner, "ema_model") and (runner.epoch >= self.ema_eval_start_epoch):
            results_ema = single_gpu_test(runner.ema_model.module, self.dataloader)
            self.evaluate(runner, results, results_ema)
        else:
            self.evaluate(runner, results)

    def after_train_epoch(self, runner):
        if not self.by_epoch or not self.every_n_epochs(runner, self.interval):
            return
        self._do_evaluate(runner, ema=True)

    def after_train_iter(self, runner):
        if self.by_epoch or not self.every_n_iters(runner, self.interval):
            return
        runner.log_buffer.clear()
        self._do_evaluate(runner)

    def evaluate(self, runner, results, results_ema=None):
        eval_res = self.dataloader.dataset.evaluate(results, logger=runner.logger, **self.eval_kwargs)
        score = eval_res[self.metric]
        for name, val in eval_res.items():
            runner.log_buffer.output[name] = val

        if results_ema:
            eval_res_ema = self.dataloader.dataset.evaluate(results_ema, logger=runner.logger, **self.eval_kwargs)
            score_ema = eval_res_ema[self.metric]
            for name, val in eval_res_ema.items():
                runner.log_buffer.output[name + "_EMA"] = val
            if score_ema > score:
                runner.save_ema_model = True

        runner.log_buffer.ready = True
        if score >= self.best_score:
            self.best_score = score
            runner.save_ckpt = True


def single_gpu_test(model, data_loader):
    model.eval()
    results = []
    dataset = data_loader.dataset
    prog_bar = mmcv.ProgressBar(len(dataset))
    for i, data in enumerate(data_loader):
        with torch.no_grad():
            result = model(return_loss=False, **data)
        results.append(result)

        batch_size = data["img"].size(0)
        for _ in range(batch_size):
            prog_bar.update()
    return results


@HOOKS.register_module()
class DistCustomEvalHook(CustomEvalHook):
    def __init__(self, dataloader, interval=1, gpu_collect=False, by_epoch=True, **eval_kwargs):
        if not isinstance(dataloader, DataLoader):
            raise TypeError("dataloader must be a pytorch DataLoader, but got " f"{type(dataloader)}")
        self.gpu_collect = gpu_collect
        super(DistCustomEvalHook, self).__init__(dataloader, interval, by_epoch=by_epoch, **eval_kwargs)

    def _do_evaluate(self, runner):
        """perform evaluation"""
        from mmcls.apis import multi_gpu_test

        results = multi_gpu_test(
            runner.model, self.dataloader, tmpdir=osp.join(runner.work_dir, ".eval_hook"), gpu_collect=self.gpu_collect
        )
        if runner.rank == 0:
            print("\n")
            self.evaluate(runner, results)

    def after_train_epoch(self, runner):
        if not self.by_epoch or not self.every_n_epochs(runner, self.interval):
            return
        self._do_evaluate(runner)

    def after_train_iter(self, runner):
        if self.by_epoch or not self.every_n_iters(runner, self.interval):
            return
        runner.log_buffer.clear()
        self._do_evaluate(runner)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field
from typing import List, Optional

from .builder import OPS
from .op import Attribute, Operation


@dataclass
class ProposalV4Attribute(Attribute):
    base_size: int
    pre_nms_topn: int
    post_nms_topn: int
    nms_thresh: float
    feat_stride: int
    min_size: int
    ratio: List[float]
    scale: List[float]
    clip_before_nms: bool = field(default=True)
    clip_after_nms: bool = field(default=False)
    normalize: bool = field(default=False)
    box_size_scale: float = field(default=1.0)
    box_coordinate_scale: float = field(default=1.0)
    framework: str = field(default="")

    def __post_init__(self):
        super().__post_init__()
        valid_framework = ["", "tensorflow"]
        if self.framework not in valid_framework:
            raise ValueError(f"Invalid framework {self.framework}. " f"It must be one of {valid_framework}.")


@OPS.register()
class ProposalV4(Operation[ProposalV4Attribute]):
    TYPE = "Proposal"
    VERSION = 4
    ATTRIBUTE_FACTORY = ProposalV4Attribute

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        #  from mmdet.core.anchor.anchor_generator import AnchorGenerator
        #  self._anchor_generator = AnchorGenerator(
        #      strides=[attrs["feat_stride"]],
        #      ratios=attrs["ratio"],
        #      scales=attrs["scale"],
        #      base_sizes=[attrs["base_size"]],
        #  )

        #  from torchvision.models.detection.anchor_utils import AnchorGenerator
        #  self._anchor_generator = AnchorGenerator(
        #      sizes=(self.attrs["base_size"],),
        #      aspect_ratios=

    def forward(self, class_probs, bbox_deltas, image_shape):
        raise NotImplementedError


@dataclass
class ROIPoolingV0Attribute(Attribute):
    pooled_h: int
    pooled_w: int
    spatial_scale: float
    method: str = field(default="max")
    output_size: List[int] = field(default_factory=lambda: [])

    def __post_init__(self):
        super().__post_init__()
        valid_method = ["max", "bilinear"]
        if self.method not in valid_method:
            raise ValueError(f"Invalid method {self.method}. " f"It must be one of {valid_method}.")


@OPS.register()
class ROIPoolingV0(Operation[ROIPoolingV0Attribute]):
    TYPE = "ROIPooling"
    VERSION = 0
    ATTRIBUTE_FACTORY = ROIPoolingV0Attribute

    def forward(self, input, boxes):
        raise NotImplementedError


@dataclass
class DetectionOutputV0Attribute(Attribute):
    keep_top_k: List[int]
    nms_threshold: float
    background_label_id: int = field(default=0)
    top_k: int = field(default=-1)
    variance_encoded_in_target: bool = field(default=False)
    code_type: str = field(default="caffe.PriorBoxParameter.CORNER")
    share_location: bool = field(default=True)
    confidence_threshold: float = field(default=0)
    clip_after_nms: bool = field(default=False)
    clip_before_nms: bool = field(default=False)
    decrease_label_id: bool = field(default=False)
    normalized: bool = field(default=False)
    input_height: int = field(default=1)
    input_width: int = field(default=1)
    objectness_score: float = field(default=0)

    def __post_init__(self):
        super().__post_init__()
        valid_code_type = [
            "caffe.PriorBoxParameter.CORNER",
            "caffe.PriorBoxParameter.CENTER_SIZE",
        ]
        if self.code_type not in valid_code_type:
            raise ValueError(f"Invalid code_type {self.code_type}. " f"It must be one of {valid_code_type}.")


@OPS.register()
class DetectionOutputV0(Operation[DetectionOutputV0Attribute]):
    TYPE = "DetectionOutput"
    VERSION = 0
    ATTRIBUTE_FACTORY = DetectionOutputV0Attribute

    def forward(self, loc_data, conf_data, prior_data, arm_conf_data=None, arm_loc_data=None):
        raise NotImplementedError


@dataclass
class RegionYoloV0Attribute(Attribute):
    axis: int
    coords: int
    classes: int
    end_axis: int
    num: int
    anchors: Optional[List[float]] = field(default=None)
    do_softmax: bool = field(default=True)
    mask: List[int] = field(default_factory=lambda: [])


@OPS.register()
class RegionYoloV0(Operation[RegionYoloV0Attribute]):
    TYPE = "RegionYolo"
    VERSION = 0
    ATTRIBUTE_FACTORY = RegionYoloV0Attribute

    def forward(self, input):
        raise NotImplementedError


@dataclass
class PriorBoxV0Attribute(Attribute):
    offset: float
    min_size: List[float] = field(default_factory=lambda: [])
    max_size: List[float] = field(default_factory=lambda: [])
    aspect_ratio: List[float] = field(default_factory=lambda: [])
    flip: bool = field(default=False)
    clip: bool = field(default=False)
    step: float = field(default=0)
    variance: List[float] = field(default_factory=lambda: [])
    scale_all_sizes: bool = field(default=True)
    fixed_ratio: List[float] = field(default_factory=lambda: [])
    fixed_size: List[float] = field(default_factory=lambda: [])
    density: List[float] = field(default_factory=lambda: [])


@OPS.register()
class PriorBoxV0(Operation[PriorBoxV0Attribute]):
    TYPE = "PriorBox"
    VERSION = 0
    ATTRIBUTE_FACTORY = PriorBoxV0Attribute

    def forward(self, output_size, image_size):
        raise NotImplementedError


@dataclass
class PriorBoxClusteredV0Attribute(Attribute):
    offset: float
    width: List[float] = field(default_factory=lambda: [1.0])
    height: List[float] = field(default_factory=lambda: [1.0])
    clip: bool = field(default=False)
    step: float = field(default=0.0)
    step_w: float = field(default=0.0)
    step_h: float = field(default=0.0)
    variance: List[float] = field(default_factory=lambda: [])


@OPS.register()
class PriorBoxClusteredV0(Operation[PriorBoxClusteredV0Attribute]):
    TYPE = "PriorBoxClustered"
    VERSION = 0
    ATTRIBUTE_FACTORY = PriorBoxClusteredV0Attribute

    def forward(self, output_size, image_size):
        raise NotImplementedError



"""EfficientNet-B0 config for semi-supervised multi-class classification."""

# pylint: disable=invalid-name

_base_ = "../../base/models/efficientnet.py"

model = dict(
    type="SemiSLClassifier",
    task="classification",
    backbone=dict(
        version="b0",
    ),
    head=dict(
        type="SemiLinearClsHead",
        loss=dict(
            type="CrossEntropyLoss",
            loss_weight=1.0,
        ),
    ),
)

fp16 = dict(loss_scale=512.0)



"""Patch mmdet."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=protected-access,redefined-outer-name

from functools import partial

from mmdet.core.bbox.assigners.base_assigner import BaseAssigner
from mmdet.core.bbox.builder import BBOX_ASSIGNERS, BBOX_SAMPLERS
from mmdet.core.bbox.samplers import BaseSampler
from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.base_dense_head import BaseDenseHead
from mmdet.models.dense_heads.base_mask_head import BaseMaskHead
from mmdet.models.detectors.base import BaseDetector
from mmdet.models.roi_heads.base_roi_head import BaseRoIHead
from mmdet.models.roi_heads.bbox_heads.bbox_head import BBoxHead
from mmdet.models.roi_heads.bbox_heads.sabl_head import SABLHead
from mmdet.models.roi_heads.mask_heads.fcn_mask_head import FCNMaskHead

from otx.algorithms.common.adapters.nncf import (
    NNCF_PATCHER,
    is_in_nncf_tracing,
    nncf_trace_wrapper,
    no_nncf_trace_wrapper,
)
from otx.algorithms.common.adapters.nncf.patches import nncf_trace_context
from otx.mpa.deploy.utils import is_mmdeploy_enabled

HEADS_TARGETS = dict(
    classes=(
        BaseDenseHead,
        BaseMaskHead,
        BaseRoIHead,
        FCNMaskHead,
        BBoxHead,
        SABLHead,
    ),
    fn_names=("loss", "onnx_export", "get_bboxes", "get_seg_masks"),
)

BBOX_ASSIGNERS_TARGETS = dict(
    classes=(BaseAssigner,),
    fn_names=("assign",),
)

SAMPLERS_TARGETS = dict(
    classes=(BaseSampler,),
    fn_names=("sample",),
)


def _should_wrap(obj_cls, fn_name, targets):
    classes = targets["classes"]
    fn_names = targets["fn_names"]

    if obj_cls is None:
        return False
    if fn_name not in fn_names:
        return False
    if issubclass(obj_cls, classes) and getattr(obj_cls, fn_name, None):
        return True
    return False


def _wrap_mmdet_head(obj_cls):
    for fn_name in HEADS_TARGETS["fn_names"]:
        if _should_wrap(obj_cls, fn_name, HEADS_TARGETS):
            NNCF_PATCHER.patch((obj_cls, fn_name), no_nncf_trace_wrapper)
            # 'onnx_export' method calls 'forward' method which need to be traced
            NNCF_PATCHER.patch((obj_cls, "forward"), nncf_trace_wrapper)


def _wrap_mmdet_bbox_assigner(obj_cls):
    for fn_name in BBOX_ASSIGNERS_TARGETS["fn_names"]:
        if _should_wrap(obj_cls, fn_name, BBOX_ASSIGNERS_TARGETS):
            NNCF_PATCHER.patch((obj_cls, fn_name), no_nncf_trace_wrapper)


def _wrap_mmdet_sampler(obj_cls):
    for fn_name in SAMPLERS_TARGETS["fn_names"]:
        if _should_wrap(obj_cls, fn_name, SAMPLERS_TARGETS):
            NNCF_PATCHER.patch((obj_cls, fn_name), no_nncf_trace_wrapper)


# pylint: disable=invalid-name,unused-argument
def _wrap_register_module(self, fn, *args, **kwargs):
    """A function to wrap classes lazily defined such as custom ones."""

    module = kwargs.get("module", args[0] if args else None)
    assert module is not None
    _wrap_mmdet_head(module)
    _wrap_mmdet_bbox_assigner(module)
    _wrap_mmdet_sampler(module)
    return fn(*args, **kwargs)


# for mmdet defined heads
for head_cls in [BaseDenseHead, BaseMaskHead, BaseRoIHead] + list(HEADS.module_dict.values()):
    _wrap_mmdet_head(head_cls)

# for mmdet defined bbox assigners
for bbox_assigner_cls in [BaseAssigner] + list(BBOX_ASSIGNERS.module_dict.values()):
    _wrap_mmdet_bbox_assigner(bbox_assigner_cls)

# for mmdet defined samplers
# NNCF can not trace this part with torch older 1.11.0
for sampler_cls in [BaseSampler] + list(BBOX_SAMPLERS.module_dict.values()):
    _wrap_mmdet_sampler(sampler_cls)

# for custom defined
NNCF_PATCHER.patch(HEADS._register_module, _wrap_register_module)
NNCF_PATCHER.patch(BBOX_ASSIGNERS._register_module, _wrap_register_module)
NNCF_PATCHER.patch(BBOX_SAMPLERS._register_module, _wrap_register_module)
NNCF_PATCHER.patch(
    "mmdet.models.roi_heads.roi_extractors.SingleRoIExtractor.map_roi_levels",
    no_nncf_trace_wrapper,
)
NNCF_PATCHER.patch("mmdet.core.bbox2result", no_nncf_trace_wrapper)
NNCF_PATCHER.patch("mmdet.core.bbox2roi", no_nncf_trace_wrapper)


def _wrap_is_in_onnx_export(ctx, fn):
    # TODO: find a better way to solve this w/o patching 'torch.onnx.is_in_onnx_export'
    #
    # prevent incomplete graph building for MaskRCNN models
    #
    # possible alternatives
    #    - take the onnx branch in all cases

    import inspect

    stack = inspect.stack()
    frame_info = stack[2]
    if (
        frame_info.function == "forward"
        and "self" in frame_info.frame.f_locals.keys()
        and frame_info.frame.f_locals["self"].__class__.__name__ == "SingleRoIExtractor"
    ):
        return fn() or is_in_nncf_tracing()
    return fn()


NNCF_PATCHER.patch("torch.onnx.is_in_onnx_export", _wrap_is_in_onnx_export)

# add nncf context method that will be used when nncf tracing
BaseDetector.nncf_trace_context = nncf_trace_context


if is_mmdeploy_enabled():
    import mmdeploy.codebase.mmdet  # noqa: F401  # pylint: disable=unused-import
    from mmdeploy.core import FUNCTION_REWRITER
    from mmdeploy.core.rewriters.rewriter_utils import import_function

    for fn_path, record_dicts in FUNCTION_REWRITER._registry._rewrite_records.items():
        if fn_path.startswith("torch"):
            continue
        obj, obj_cls = import_function(fn_path)
        fn_name = fn_path.split(".")[-1]
        if _should_wrap(obj_cls, fn_name, HEADS_TARGETS):
            for record_dict in record_dicts:
                record_dict["_object"] = partial(no_nncf_trace_wrapper, None, record_dict["_object"])



"""This file defines the ModelConfiguration, ModelEntity and Model classes."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import copy
import os
from dataclasses import dataclass, field
from enum import Enum, IntEnum, auto
from typing import Dict, List, NamedTuple, Optional, Sequence, Union, cast

from omegaconf import DictConfig, ListConfig, OmegaConf

from otx.api.configuration.elements import metadata_keys
from otx.api.configuration.enums import AutoHPOState
from otx.api.configuration.helper.utils import search_in_config_dict
from otx.api.entities.label import Domain


class TargetDevice(IntEnum):
    """Represents the target device for a given model.

    This device might be used for instance be used for training or inference.
    """

    UNSPECIFIED = auto()
    CPU = auto()
    GPU = auto()
    VPU = auto()


class ModelOptimizationMethod(Enum):
    """Optimized model format."""

    TENSORRT = auto()
    OPENVINO = auto()

    def __str__(self) -> str:
        """Returns ModelOptimizationMethod as string."""
        return str(self.name)


@dataclass
class DatasetRequirements:
    """Expected requirements for the dataset in order to use this algorithm.

    Attributes:
        classes (Optional[List[str]]): Classes which must be present in the dataset
    """

    classes: Optional[List[str]] = None


@dataclass
class ExportableCodePaths:
    """The paths to the different versions of the exportable code for a given model template."""

    default: Optional[str] = None
    openvino: Optional[str] = None


class TaskFamily(Enum):
    """Overall task family."""

    VISION = auto()
    FLOW_CONTROL = auto()
    DATASET = auto()

    def __str__(self) -> str:
        """Returns task family as a string."""
        return str(self.name)


class TaskInfo(NamedTuple):
    """Task information.

    NamedTuple to store information about the task type like label domain, if it is
    trainable, if it is an anomaly task and if it supports global or local labels.
    """

    domain: Domain
    is_trainable: bool
    is_anomaly: bool
    is_global: bool
    is_local: bool


class TaskType(Enum):
    """The type of algorithm within the task family.

    Also contains relevant information about the task type like label domain, if it is trainable,
    if it is an anomaly task or if it supports global or local labels.

    Args:
        value (int): (Unused) Unique integer for .value property of Enum (auto() does not work)
        task_info (TaskInfo): NamedTuple containing information about the task's capabilities
    """

    # pylint: disable=unused-argument
    def __init__(
        self,
        value: int,
        task_info: TaskInfo,
    ):
        self.domain = task_info.domain
        self.is_trainable = task_info.is_trainable
        self.is_anomaly = task_info.is_anomaly
        self.is_global = task_info.is_global
        self.is_local = task_info.is_local

    def __new__(cls, *args):
        """Returns new instance."""
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    NULL = 1, TaskInfo(
        domain=Domain.NULL,
        is_trainable=False,
        is_anomaly=False,
        is_global=False,
        is_local=False,
    )
    DATASET = 2, TaskInfo(
        domain=Domain.NULL,
        is_trainable=False,
        is_anomaly=False,
        is_global=False,
        is_local=False,
    )
    CLASSIFICATION = 3, TaskInfo(
        domain=Domain.CLASSIFICATION,
        is_trainable=True,
        is_anomaly=False,
        is_global=True,
        is_local=False,
    )
    SEGMENTATION = 4, TaskInfo(
        domain=Domain.SEGMENTATION,
        is_trainable=True,
        is_anomaly=False,
        is_global=False,
        is_local=True,
    )
    DETECTION = 5, TaskInfo(
        domain=Domain.DETECTION,
        is_trainable=True,
        is_anomaly=False,
        is_global=False,
        is_local=True,
    )
    ANOMALY_DETECTION = 6, TaskInfo(
        domain=Domain.ANOMALY_DETECTION,
        is_trainable=True,
        is_anomaly=True,
        is_global=False,
        is_local=True,
    )
    CROP = 7, TaskInfo(
        domain=Domain.NULL,
        is_trainable=False,
        is_anomaly=False,
        is_global=False,
        is_local=False,
    )
    TILE = 8, TaskInfo(
        domain=Domain.NULL,
        is_trainable=False,
        is_anomaly=False,
        is_global=False,
        is_local=False,
    )
    INSTANCE_SEGMENTATION = 9, TaskInfo(
        domain=Domain.INSTANCE_SEGMENTATION,
        is_trainable=True,
        is_anomaly=False,
        is_global=False,
        is_local=True,
    )
    ACTIVELEARNING = 10, TaskInfo(
        domain=Domain.NULL,
        is_trainable=False,
        is_anomaly=False,
        is_global=False,
        is_local=False,
    )
    ANOMALY_SEGMENTATION = 11, TaskInfo(
        domain=Domain.ANOMALY_SEGMENTATION,
        is_trainable=True,
        is_anomaly=True,
        is_global=False,
        is_local=True,
    )
    ANOMALY_CLASSIFICATION = 12, TaskInfo(
        domain=Domain.ANOMALY_CLASSIFICATION,
        is_trainable=True,
        is_anomaly=True,
        is_global=True,
        is_local=False,
    )
    ROTATED_DETECTION = 13, TaskInfo(
        domain=Domain.ROTATED_DETECTION,
        is_trainable=True,
        is_anomaly=False,
        is_global=False,
        is_local=True,
    )
    ACTION_CLASSIFICATION = 14, TaskInfo(
        domain=Domain.ACTION_CLASSIFICATION,
        is_trainable=True,
        is_anomaly=False,
        is_global=False,
        is_local=True,
    )
    ACTION_DETECTION = 15, TaskInfo(
        domain=Domain.ACTION_DETECTION, is_trainable=True, is_anomaly=False, is_global=False, is_local=True
    )

    def __str__(self) -> str:
        """Returns name."""
        return self.name

    def __repr__(self) -> str:
        """Returns name."""
        return self.name


def task_type_to_label_domain(task_type: TaskType) -> Domain:
    """Links the task type to the label domain enum.

    Note that not all task types have an associated domain (e.g. crop task).
    In this case, a ``ValueError`` is raised.

    Args:
        task_type (TaskType): The task type to get the label domain for.

    Returns:
        Domain: The label domain for the task type.
    """
    mapping = {
        TaskType.CLASSIFICATION: Domain.CLASSIFICATION,
        TaskType.DETECTION: Domain.DETECTION,
        TaskType.SEGMENTATION: Domain.SEGMENTATION,
        TaskType.INSTANCE_SEGMENTATION: Domain.INSTANCE_SEGMENTATION,
        TaskType.ANOMALY_CLASSIFICATION: Domain.ANOMALY_CLASSIFICATION,
        TaskType.ANOMALY_DETECTION: Domain.ANOMALY_DETECTION,
        TaskType.ANOMALY_SEGMENTATION: Domain.ANOMALY_SEGMENTATION,
        TaskType.ROTATED_DETECTION: Domain.ROTATED_DETECTION,
        TaskType.ACTION_CLASSIFICATION: Domain.ACTION_CLASSIFICATION,
        TaskType.ACTION_DETECTION: Domain.ACTION_DETECTION,
    }

    try:
        return mapping[task_type]
    except KeyError as exc:
        raise ValueError(f"Task type {task_type} does not have any associated label domain.") from exc


@dataclass
class HyperParameterData:
    """HyperParameter Data.

    Class that contains the raw hyper parameter data, for those hyper parameters for the model that are
    user-configurable.

    Attributes:
        base_path (Optional[str]): The path to the yaml file specifying the base configurable parameters to use in the
            model. Defaults to None.
        parameter_overrides (Dict): Nested dictionary that describes overrides for the metadata for the
            user-configurable hyper parameters that are used in the model. This allows multiple models to share the
            same base hyper-parameters, while for each individual model the defaults, parameter ranges, descriptions,
            etc. can still be customized.
    """

    base_path: Optional[str] = None
    parameter_overrides: Dict = field(default_factory=dict)
    __data: Dict = field(default_factory=dict, repr=False)
    __has_valid_configurable_parameters: bool = field(default=False, repr=False)

    def load_parameters(self, model_template_path: str):
        """Load hyper parameters.

        Loads the actual hyper parameters defined in the file at `base_path`, and performs any overrides specified in
        the `parameter_overrides`.

        Args:
            model_template_path (str): file path to the model template file in which the HyperParameters live.
        """
        has_valid_configurable_parameters = False
        if self.base_path is not None and os.path.exists(model_template_path):
            model_template_dir = os.path.dirname(model_template_path)
            base_hyper_parameter_path = os.path.join(model_template_dir, self.base_path)

            config_dict = OmegaConf.load(base_hyper_parameter_path)
            data = OmegaConf.to_container(config_dict)
            if isinstance(data, dict):
                self.__remove_parameter_values_from_data(data)
                self.__data = data
                has_valid_configurable_parameters = True
            else:
                raise ValueError(
                    f"Unexpected configurable parameter file found at path {base_hyper_parameter_path}"
                    f", expected a dictionary-like format, got list-like instead."
                )
        if self.has_overrides and has_valid_configurable_parameters:
            self.substitute_parameter_overrides()
        self.__has_valid_configurable_parameters = has_valid_configurable_parameters

    @property
    def data(self) -> Dict:
        """Returns a dictionary containing the set of hyper parameters defined in the ModelTemplate.

        This does not contain the actual parameter values, but instead holds the parameter schema's in
        a structured manner. The actual values should be either loaded from the database, or will be initialized from
        the defaults upon creating a configurable parameter object out of this data.
        """
        return self.__data

    @property
    def has_overrides(self) -> bool:
        """Returns True if any parameter overrides are defined by the HyperParameters instance, False otherwise."""
        return self.parameter_overrides != {}

    @property
    def has_valid_configurable_parameters(self) -> bool:
        """Check if configurable parameters are valid.

        Returns True if the HyperParameterData instance contains valid configurable parameters, extracted from the
        model template. False otherwise.
        """
        return self.__has_valid_configurable_parameters

    def substitute_parameter_overrides(self):
        """Carries out the parameter overrides specified in the `parameter_overrides` attribute.

        Validates whether the overridden parameters exist in the base set of configurable parameters,
        and whether the metadata values that should be overridden are valid metadata attributes.
        """
        self.__substitute_parameter_overrides(self.parameter_overrides, self.__data)

    def __substitute_parameter_overrides(self, override_dict: Dict, parameter_dict: Dict):
        """Substitutes parameters form override_dict into parameter_dict.

        Recursively substitutes overridden parameter values specified in `override_dict` into the base set of
        hyper parameters passed in as `parameter_dict`

        Args:
            override_dict (Dict): dictionary containing the parameter overrides
            parameter_dict (Dict): dictionary that contains the base set of hyper parameters, in which the overridden
                values are substituted
        """
        for key, value in override_dict.items():
            if isinstance(value, dict) and not metadata_keys.allows_dictionary_values(key):
                if key in parameter_dict.keys():
                    self.__substitute_parameter_overrides(value, parameter_dict[key])
                else:
                    raise ValueError(
                        f"Unable to perform parameter override. Parameter or parameter group named {key} "
                        f"is not valid for the base hyper parameters specified in {self.base_path}"
                    )
            else:
                if metadata_keys.allows_model_template_override(key):
                    parameter_dict[key] = value
                else:
                    raise KeyError(f"{key} is not a valid keyword for hyper parameter overrides")

    @classmethod
    def __remove_parameter_values_from_data(cls, data: dict):
        """This method removes the actual parameter values from the input parameter data.

        These values should be removed because the parameters should be instantiated
        from the default_values, instead of their values.

        NOTE: This method modifies its input dictionary, it does not return a new copy

        Args:
            data: Parameter dictionary to remove values from
        """
        data_copy = copy.deepcopy(data)
        for key, value in data_copy.items():
            if isinstance(value, dict):
                if key != metadata_keys.UI_RULES:
                    cls.__remove_parameter_values_from_data(data[key])
            elif key == "value":
                data.pop(key)

    def manually_set_data_and_validate(self, hyper_parameters: dict):
        """This function is used to manually set the hyper parameter data from a dictionary.

        It is meant to be used in testing only, in cases where the model
        template is not backed up by an actual yaml file.

        Args:
            hyper_parameters (Dict): Dictionary containing the data to be set
        """
        self.__data = hyper_parameters
        self.__has_valid_configurable_parameters = True


class InstantiationType(Enum):
    """The method to instantiate a given task."""

    NONE = auto()
    CLASS = auto()
    GRPC = auto()

    def __str__(self) -> str:
        """Returns the name of the instantiation type."""
        return str(self.name)


@dataclass
class Dependency:
    """Dependency required by the task.

    Attributes:
        source (str): Source of the dependency
        destination (str): Destination folder to install the dependency
        size (Optional[int]): Size of the dependency in bytes
        sha256 (Optional[str]): SHA-256 checksum of the dependency file
    """

    source: str
    destination: str
    size: Optional[int] = None
    sha256: Optional[str] = None


@dataclass
class EntryPoints:
    """Path of the Python classes implementing the task interface.

    Attributes:
        base (str): Base interface implementing the functionality in a framework such as PyTorch or TensorFlow
        openvino (Optional[str]): OpenVINO interface.
        nncf (Optional[str]): NNCF interface
    """

    base: str
    openvino: Optional[str] = None
    nncf: Optional[str] = None


# pylint: disable=too-many-instance-attributes
@dataclass
class ModelTemplate:
    """This class represents a Task in the Task database.

    It can be either a CLASS type, with the class path specified or a GRPC type with its address.
    The task chain uses this information to setup a `ChainLink` (A task in the chain)

    model_template_id (str): ID of the model template
    model_template_path (str): path to the original model template file
    name (str): user-friendly name for the algorithm used in the task
    task_family (TaskFamily): overall task family of the task. One of VISION, FLOW_CONTROL AND DATASET.
    task_type (TaskType): Type of algorithm within task family.
    instantiation (InstantiationType): InstantiationType (CLASS or GRPC)
    summary (str): Summary of what the algorithm does. Defaults to "".
    framework (Optional[str]): The framework used by the algorithm. Defaults to None.
    max_nodes (int): Max number of nodes for training. Defaults to 1.
    application (Optional[str]): Name of the application solved by this algorithm. Defaults to None.
    dependencies (Liar[Dependency]): List of dependencies required by the algorithm. Defaults to empty `field`.
    initial_weights (Optional[str]): Optional URL to the initial weights used by the algorithm. Defaults to None
    training_targets (List[TargetDevice]): device used for training. Defaults to empty `field`.
    inference_targets (List[TargetDevices]): device used for inference. Defaults to empty `field`.
    dataset_requirements (DatasetRequirements): list of dataset requirements. Defaults to empty `field`.
    model_optimization_methods (List[ModelOptimizationMethod]): list of ModelOptimizationMethod.
        This lists all methods available to optimize the inference model for the task
    hyper_parameters (HyperParameterData): HyperParameterData object containing the base path to the configurable
        parameter definition, as well as any overrides for the base parameters that are specific for the
        current template.
    is_trainable (bool): specify whether task is trainable
    capabilities (List[str]): list of task capabilities
    grpc_address (Optional[str]): the grpc host address (for instantiation type == GRPC)
    entrypoints (Optional[Entrypoints]): Entrypoints implementing the Python task interface
    base_model_path (str): Path to template file for the base model used for nncf compression.
    exportable_code_paths (ExportableCodePaths): if it exists, the path to the exportable code sources.
        Defaults to empty `field`.
    task_type_sort_priority (int): priority of order of how tasks are shown in the pipeline dropdown for a given task
        type. E.g. for classification Inception is default and has weight 0. Unassigned priority will have -1 as
        priority. mobilenet is less important, and has a higher value. Default is zero (the highest priority).
    gigaflops (float): how many billions of operations are required to do inference on a single data item.
    size (float): how much disk space the model will approximately take.
    """

    model_template_id: str
    model_template_path: str
    name: str
    task_family: TaskFamily
    task_type: TaskType
    instantiation: InstantiationType
    summary: str = ""
    framework: Optional[str] = None
    max_nodes: int = 1
    application: Optional[str] = None
    dependencies: List[Dependency] = field(default_factory=list)
    initial_weights: Optional[str] = None
    training_targets: List[TargetDevice] = field(default_factory=list)
    inference_targets: List[TargetDevice] = field(default_factory=list)
    dataset_requirements: DatasetRequirements = field(default_factory=DatasetRequirements)
    model_optimization_methods: List[ModelOptimizationMethod] = field(default_factory=list)
    hyper_parameters: HyperParameterData = field(default_factory=HyperParameterData)
    is_trainable: bool = True
    capabilities: List[str] = field(default_factory=list)
    grpc_address: Optional[str] = None
    entrypoints: Optional[EntryPoints] = None
    base_model_path: str = ""
    exportable_code_paths: ExportableCodePaths = field(default_factory=ExportableCodePaths)
    task_type_sort_priority: int = -1
    gigaflops: float = 0
    size: float = 0
    hpo: Optional[Dict] = None

    def __post_init__(self):
        """Do sanitation checks before loading the hyper-parameters."""
        if self.instantiation == InstantiationType.GRPC and self.grpc_address == "":
            raise ValueError("Task is registered as gRPC, but no gRPC address is specified")
        if self.instantiation == InstantiationType.CLASS and self.entrypoints is None:
            raise ValueError("Task is registered as CLASS, but entrypoints were not specified")
        if self.task_family == TaskFamily.VISION and self.hyper_parameters.base_path is None:
            raise ValueError("Task is registered as a VISION task but no hyper parameters were defined.")
        if self.task_family != TaskFamily.VISION and self.hyper_parameters.base_path is not None:
            raise ValueError("Hyper parameters are currently not supported for non-VISION tasks.")

        # Load the full hyper parameters
        self.hyper_parameters.load_parameters(self.model_template_path)

    def computes_uncertainty_score(self) -> bool:
        """Returns true if "compute_uncertainty_score" is in capabilities false otherwise."""
        return "compute_uncertainty_score" in self.capabilities

    def computes_representations(self) -> bool:
        """Returns true if "compute_representations" is in capabilities."""
        return "compute_representations" in self.capabilities

    def is_task_global(self) -> bool:
        """Returns ``True`` if the task is global task i.e. if task produces global labels."""
        return self.task_type.is_global

    def supports_auto_hpo(self) -> bool:
        """Returns `True` if the algorithm supports automatic hyper parameter optimization, `False` otherwise."""
        if not self.hyper_parameters.has_valid_configurable_parameters:
            return False
        auto_hpo_state_results = search_in_config_dict(
            self.hyper_parameters.data, key_to_search=metadata_keys.AUTO_HPO_STATE
        )
        for result in auto_hpo_state_results:
            if str(result[0]).lower() == str(AutoHPOState.POSSIBLE):
                return True
        return False


class NullModelTemplate(ModelTemplate):
    """Represent an empty model template. Note that a task based on this model template cannot be instantiated."""

    def __init__(self) -> None:
        super().__init__(
            model_template_id="",
            model_template_path="",
            task_family=TaskFamily.FLOW_CONTROL,
            task_type=TaskType.NULL,
            name="Null algorithm",
            instantiation=InstantiationType.NONE,
            capabilities=[],
        )


ANOMALY_TASK_TYPES: Sequence[TaskType] = (
    TaskType.ANOMALY_DETECTION,
    TaskType.ANOMALY_CLASSIFICATION,
    TaskType.ANOMALY_SEGMENTATION,
)


TRAINABLE_TASK_TYPES: Sequence[TaskType] = (
    TaskType.CLASSIFICATION,
    TaskType.DETECTION,
    TaskType.SEGMENTATION,
    TaskType.INSTANCE_SEGMENTATION,
    TaskType.ANOMALY_DETECTION,
    TaskType.ANOMALY_CLASSIFICATION,
    TaskType.ANOMALY_SEGMENTATION,
    TaskType.ROTATED_DETECTION,
    TaskType.ACTION_CLASSIFICATION,
    TaskType.ACTION_DETECTION,
)


def _parse_model_template_from_omegaconf(config: Union[DictConfig, ListConfig]) -> ModelTemplate:
    """Parse an OmegaConf configuration into a model template.

    Args:
        config (Union[DictConfig, ListConfig]): The configuration to parse.

    Returns:
        ModelTemplate: The parsed model template.
    """
    schema = OmegaConf.structured(ModelTemplate)
    config = OmegaConf.merge(schema, config)
    return cast(ModelTemplate, OmegaConf.to_object(config))


def parse_model_template(model_template_path: str) -> ModelTemplate:
    """Read a model template from a file.

    Args:
        model_template_path (str): Path to the model template template.yaml file

    Returns:
        ModelTemplate: The model template parsed from the file.
    """
    config = OmegaConf.load(model_template_path)
    if not isinstance(config, DictConfig):
        raise ValueError("Expected the configuration file to contain a dictionary, not a list")

    if "model_template_id" not in config:
        config["model_template_id"] = config["name"].replace(" ", "_")
    config["model_template_path"] = model_template_path
    return _parse_model_template_from_omegaconf(config)


def parse_model_template_from_dict(model_template_dict: dict) -> ModelTemplate:
    """Read a model template from a dictionary.

    Note that the model_template_id must be defined inside the dictionary.

    Args:
        model_template_dict (dict): Dictionary containing the model template.

    Returns:
        ModelTemplate: The model template.
    """
    config = OmegaConf.create(model_template_dict)
    return _parse_model_template_from_omegaconf(config)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmseg.models.segmentors.base import BaseSegmentor

from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_patches():
    import otx.algorithms.segmentation.adapters.mmseg.nncf.patches  # noqa: F401

    assert getattr(BaseSegmentor, "nncf_trace_context", None) is not None



# Copyright (c) 2019 MendelXu
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn


class PSPModule(nn.Module):
    """Reference: https://github.com/MendelXu/ANN"""

    methods = {"max": nn.AdaptiveMaxPool2d, "avg": nn.AdaptiveAvgPool2d}

    def __init__(self, sizes=(1, 3, 6, 8), method="max"):
        super().__init__()

        assert method in self.methods
        pool_block = self.methods[method]

        self.stages = nn.ModuleList([pool_block(output_size=(size, size)) for size in sizes])

    def forward(self, feats):
        n, c, _, _ = feats.size()

        priors = [stage(feats).view(n, c, -1) for stage in self.stages]
        out = torch.cat(priors, -1)

        return out



"""OTX Adapters - mmdet."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from .data import OTXDetDataset

# fmt: off
# isort: off
# FIXME: openvino pot library adds stream handlers to root logger
# which makes annoying duplicated logging
from mmdet.utils import get_root_logger  # pylint: disable=wrong-import-order
get_root_logger().propagate = False  # pylint: disable=wrong-import-order
# isort:on
# fmt: on

__all__ = ["OTXDetDataset"]



"""Backbone config of OCR-Lite-HRnet-18."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

model = dict(
    backbone=dict(
        type="LiteHRNet",
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=False,
        extra=dict(
            stem=dict(
                stem_channels=32,
                out_channels=32,
                expand_ratio=1,
                strides=(2, 2),
                extra_stride=False,
                input_norm=False,
            ),
            num_stages=3,
            stages_spec=dict(
                num_modules=(2, 4, 2),
                num_branches=(2, 3, 4),
                num_blocks=(2, 2, 2),
                module_type=("LITE", "LITE", "LITE"),
                with_fuse=(True, True, True),
                reduce_ratios=(8, 8, 8),
                num_channels=(
                    (40, 80),
                    (40, 80, 160),
                    (40, 80, 160, 320),
                ),
            ),
            out_modules=dict(
                conv=dict(enable=False, channels=320),
                position_att=dict(
                    enable=False,
                    key_channels=128,
                    value_channels=320,
                    psp_size=(1, 3, 6, 8),
                ),
                local_att=dict(enable=False),
            ),
            out_aggregator=dict(enable=False),
            add_input=False,
        ),
    ),
)



"""EfficientNet-V2 for multi-label config."""

# pylint: disable=invalid-name

_base_ = "../base/models/efficientnet_v2.py"

model = dict(
    type="SAMImageClassifier",
    task="classification",
    backbone=dict(
        version="s_21k",
    ),
    head=dict(
        type="CustomMultiLabelLinearClsHead",
        normalized=True,
        scale=7.0,
        loss=dict(type="AsymmetricAngularLossWithIgnore", gamma_pos=0.0, gamma_neg=1.0, reduction="sum"),
    ),
)

fp16 = dict(loss_scale=512.0)



"""Model optimization tool."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json

from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from otx.cli.manager import ConfigManager
from otx.cli.utils.importing import get_impl_class
from otx.cli.utils.io import read_model, save_model_data
from otx.cli.utils.parser import (
    add_hyper_parameters_sub_parser,
    get_parser_and_hprams_data,
)
from otx.core.data.adapter import get_dataset_adapter

# pylint: disable=too-many-locals


def get_args():
    """Parses command line arguments.

    It dynamically generates help for hyper-parameters which are specific to particular model template.
    """
    parser, hyper_parameters, params = get_parser_and_hprams_data()

    parser.add_argument(
        "--train-data-roots",
        help="Comma-separated paths to training data folders.",
    )
    parser.add_argument(
        "--val-data-roots",
        help="Comma-separated paths to validation data folders.",
    )
    parser.add_argument(
        "--load-weights",
        help="Load weights of trained model",
    )
    parser.add_argument(
        "--save-model-to",
        help="Location where trained model will be stored.",
    )
    parser.add_argument(
        "--save-performance",
        help="Path to a json file where computed performance will be stored.",
    )
    parser.add_argument(
        "--work-dir",
        help="Location where the intermediate output of the task will be stored.",
        default=None,
    )

    add_hyper_parameters_sub_parser(parser, hyper_parameters)
    override_param = [f"params.{param[2:].split('=')[0]}" for param in params if param.startswith("--")]

    return parser.parse_args(), override_param


def main():
    """Main function that is used for model training."""

    # Dynamically create an argument parser based on override parameters.
    args, override_param = get_args()

    config_manager = ConfigManager(args, workspace_root=args.work_dir, mode="train")
    # Auto-Configuration for model template
    config_manager.configure_template()

    # The default in the workspace is the model weight of the OTX train.
    if not args.load_weights and config_manager.check_workspace():
        args.load_weights = str(config_manager.workspace_root / "models/weights.pth")

    is_pot = False
    if args.load_weights.endswith(".bin") or args.load_weights.endswith(".xml"):
        is_pot = True

    template = config_manager.template
    if not is_pot and template.entrypoints.nncf is None:
        raise RuntimeError(f"Optimization by NNCF is not available for template {args.template}")

    # Update Hyper Parameter Configs
    hyper_parameters = config_manager.get_hyparams_config(override_param)

    # Get classes for Task, ConfigurableParameters and Dataset.
    task_class = get_impl_class(template.entrypoints.openvino if is_pot else template.entrypoints.nncf)

    # Auto-Configuration for Dataset configuration
    config_manager.configure_data_config(update_data_yaml=config_manager.check_workspace())
    dataset_config = config_manager.get_dataset_config(subsets=["train", "val"])
    dataset_adapter = get_dataset_adapter(**dataset_config)
    dataset, label_schema = dataset_adapter.get_otx_dataset(), dataset_adapter.get_label_schema()

    environment = TaskEnvironment(
        model=None,
        hyper_parameters=hyper_parameters,
        label_schema=label_schema,
        model_template=template,
    )

    environment.model = read_model(environment.get_model_configuration(), args.load_weights, None)

    task = task_class(task_environment=environment)

    output_model = ModelEntity(dataset, environment.get_model_configuration())

    task.optimize(
        OptimizationType.POT if is_pot else OptimizationType.NNCF,
        dataset,
        output_model,
        OptimizationParameters(),
    )

    if "save_model_to" not in args or not args.save_model_to:
        args.save_model_to = str(config_manager.workspace_root / "model-optimized")
    save_model_data(output_model, args.save_model_to)

    validation_dataset = dataset.get_subset(Subset.VALIDATION)
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(),
        InferenceParameters(is_evaluation=True),
    )

    resultset = ResultSetEntity(
        model=output_model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    task.evaluate(resultset)
    assert resultset.performance is not None
    print(resultset.performance)

    if args.save_performance:
        with open(args.save_performance, "w", encoding="UTF-8") as write_file:
            json.dump(
                {resultset.performance.score.name: resultset.performance.score.value},
                write_file,
            )

    return dict(retcode=0, template=template.name)


if __name__ == "__main__":
    main()



"""This module implements the dataset item entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=cyclic-import

import abc
import copy
import itertools
import logging
from threading import Lock
from typing import List, Optional, Sequence, Set, Tuple, Union

import numpy as np

from otx.api.entities.annotation import Annotation, AnnotationSceneEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.media import IMedia2DEntity
from otx.api.entities.metadata import IMetadata, MetadataItemEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.utils.shape_factory import ShapeFactory

logger = logging.getLogger(__name__)


class DatasetItemEntity(metaclass=abc.ABCMeta):
    """DatasetItemEntity represents an item in the DatasetEntity.

    It holds a media item, annotation and an ROI. The ROI determines the region of interest for the dataset item, and
    is described by a shape entity.

    The fundamental properties of a dataset item are:

    - A 2d media entity (e.g. Image)
    - A 2d annotation entity for the full resolution media entity
    - An ROI, describing the region of interest.
    - The subset it belongs to
    - Metadata for the media entity (e.g. saliency map or active score)
    - A list of labels to ignore

    .. rubric:: Getting data from dataset item

    The first step is to fetch the input data for the network.

    >>> dataset_item = DatasetItemEntity()
    >>> media_numpy = dataset_item.numpy  # RGB media data (Height, Width, Channels)

    This returns the numpy data for the assigned ROI. But it is possible to extract any arbitrary region.

    >>> from otx.api.entities.shapes.rectangle import Rectangle
    >>> top_left_quart_roi = Annotation(Rectangle(x1=0.0, y1=0.0, x2=0.5, y2=0.5), labels=[])
    >>> top_left_quart_numpy = dataset_item.roi_numpy(roi=top_left_quart_roi)

    Get the subset of labels for the item ROI:

    >>> labels = dataset_item.get_roi_labels(labels=...)

    Get the annotations __visible__ in the ROI:

    >>> dataset_item.get_annotations()

    .. rubric:: Adding output data to dataset item

    It is possible to add shapes or just labels for the ROI.

    Add shapes to dataset item:

    >>> box = Rectangle(x1=0.2, y1=0.3, x2=0.6, y2=0.5)
    >>> dataset_item.append_annotations(annotations=[Annotation(box, labels=[...])])

    Add labels to ROI:

    >>> dataset_item.append_labels(labels=[...])

    Args:
        media (IMedia2DEntity): Media item
        annotation_scene (AnnotationSceneEntity): Annotation scene
        roi (Optional[Annotation]): Region Of Interest
        metadata (Optional[List[MetadataItemEntity]]): Metadata attached to dataset item
        subset (Subset): `Subset` for item. E.g. `Subset.VALIDATION`
        ignored_labels (Optional[Union[List[LabelEntity], Tuple[LabelEntity, ...], Set[LabelEntity]]]): Collection of
            labels that should be ignored in this dataset item. For instance, in a training scenario, this parameter is
            used to ignore certain labels within the existing annotations because their status becomes uncertain
            following a label schema change.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        media: IMedia2DEntity,
        annotation_scene: AnnotationSceneEntity,
        roi: Optional[Annotation] = None,
        metadata: Optional[List[MetadataItemEntity]] = None,
        subset: Subset = Subset.NONE,
        ignored_labels: Optional[Union[List[LabelEntity], Tuple[LabelEntity, ...], Set[LabelEntity]]] = None,
    ):
        self.__media: IMedia2DEntity = media
        self.__annotation_scene: AnnotationSceneEntity = annotation_scene
        self.__subset: Subset = subset
        self.__roi_lock = Lock()

        # set ROI
        if roi is None:
            for annotation in annotation_scene.annotations:
                # if there is a full box in annotation.shapes, set it as ROI
                if Rectangle.is_full_box(annotation.shape):
                    roi = annotation
                    break
        self.__roi = roi

        self.__metadata: List[MetadataItemEntity] = []
        if metadata is not None:
            self.__metadata = metadata

        self.__ignored_labels: Set[LabelEntity] = set() if ignored_labels is None else set(ignored_labels)

    def set_metadata(self, metadata: List[MetadataItemEntity]):
        """Sets the metadata."""
        self.__metadata = metadata

    def get_metadata(self) -> List[MetadataItemEntity]:
        """Returns the metadata."""
        return self.__metadata

    @property
    def ignored_labels(self) -> Set[LabelEntity]:
        """Get the IDs of the labels to ignore in this dataset item."""
        return self.__ignored_labels

    @ignored_labels.setter
    def ignored_labels(self, value: Union[List[LabelEntity], Tuple[LabelEntity, ...], Set[LabelEntity]]):
        self.__ignored_labels = set(value)

    def __repr__(self):
        """String representation of the dataset item."""
        return (
            f"{self.__class__.__name__}("
            f"media={self.media}, "
            f"annotation_scene={self.annotation_scene}, "
            f"roi={self.roi}, "
            f"subset={self.subset}), "
            f"meta={self.get_metadata()}"
        )

    @property
    def roi(self) -> Annotation:
        """Region Of Interest."""
        with self.__roi_lock:
            if self.__roi is None:
                requested_roi = Annotation(Rectangle.generate_full_box(), labels=[])
                self.__roi = requested_roi
            else:
                requested_roi = self.__roi
            return requested_roi

    @roi.setter
    def roi(self, roi: Optional[Annotation]):
        with self.__roi_lock:
            self.__roi = roi

    @property
    def subset(self) -> Subset:
        """Returns the subset that the IDatasetItem belongs to. e.g. Subset.TRAINING."""
        return self.__subset

    @subset.setter
    def subset(self, value: Subset):
        self.__subset = value

    @property
    def media(self) -> IMedia2DEntity:
        """Media."""
        return self.__media

    def roi_numpy(self, roi: Optional[Annotation] = None) -> np.ndarray:
        """Gives the numpy data for the media, given an ROI.

        This function allows to take a crop of any arbitrary region of the media in the Dataset entity.
        If the ROI is not given, the ROI assigned to the DatasetItem will be used as default.

        Args:
            roi (Optional[Annotation]): Shape entity. The shape will be converted if needed, to extract the ROI numpy.

        Returns:
            np.ndarray: Numpy array with media data
        """
        if roi is None:
            roi = self.roi

        if roi is not None:
            roi.shape = ShapeFactory.shape_as_rectangle(roi.shape)

        return self.media.roi_numpy(roi=roi)

    @property
    def numpy(self) -> np.ndarray:
        """Returns the numpy data for the media, taking ROI into account.

        Returns:
            np.ndarrray: Numpy array. RGB array of shape (Height, Width, Channels)
        """
        return self.roi_numpy()

    @property
    def width(self) -> int:
        """The width of the dataset item, taking into account the ROI."""
        roi_shape_as_box = ShapeFactory.shape_as_rectangle(self.roi.shape)
        roi_shape_as_box = roi_shape_as_box.clip_to_visible_region()
        width = self.media.width

        # Note that we cannot directly use roi_shape_as_box.width due to the rounding
        # because round(x2 - x1) is not always equal to round(x2) - round(x1)
        x1 = int(round(roi_shape_as_box.x1 * width))
        x2 = int(round(roi_shape_as_box.x2 * width))
        return x2 - x1

    @property
    def height(self) -> int:
        """The height of the dataset item, taking into account the ROI."""
        roi_shape_as_box = ShapeFactory.shape_as_rectangle(self.roi.shape)
        roi_shape_as_box = roi_shape_as_box.clip_to_visible_region()
        height = self.media.height

        # Note that we cannot directly use roi_shape_as_box.height due to the rounding
        # because round(y2 - y1) is not always equal to round(y2) - round(y1)
        y1 = int(round(roi_shape_as_box.y1 * height))
        y2 = int(round(roi_shape_as_box.y2 * height))
        return y2 - y1

    @property
    def annotation_scene(self) -> AnnotationSceneEntity:
        """Access to annotation scene."""
        return self.__annotation_scene

    @annotation_scene.setter
    def annotation_scene(self, value: AnnotationSceneEntity):
        self.__annotation_scene = value

    def get_annotations(
        self,
        labels: Optional[List[LabelEntity]] = None,
        include_empty: bool = False,
        include_ignored: bool = False,
    ) -> List[Annotation]:
        """Returns a list of annotations that exist in the dataset item (wrt. ROI).

        This is done by checking that the center of the annotation is located in the ROI.

        Args:
            labels (Optional[LabelEntity]): Subset of input labels to filter with; if ``None``, all the shapes within
                the ROI are returned.
            include_empty (bool): if True, returns both empty and non-empty labels
            include_ignored (bool): if True, includes the labels in ignored_labels

        Returns:
            List[Annotation]: The intersection of the input label set and those present within the ROI
        """
        is_full_box = Rectangle.is_full_box(self.roi.shape)
        annotations = []
        if is_full_box and labels is None and include_empty and include_ignored:
            # Fast path for the case where we do not need to change the shapes
            annotations = self.annotation_scene.annotations
        else:
            # Todo: improve speed. This is O(n) for n shapes.
            roi_as_box = ShapeFactory.shape_as_rectangle(self.roi.shape)

            labels_set = {label.name for label in labels} if labels is not None else set()

            for annotation in self.annotation_scene.annotations:
                if not is_full_box and not self.roi.shape.contains_center(annotation.shape):
                    continue

                shape_labels = annotation.get_labels(include_empty)

                check_labels = False
                if not include_ignored:
                    shape_labels = [label for label in shape_labels if label.label not in self.ignored_labels]
                    check_labels = True

                if labels is not None:
                    shape_labels = [label for label in shape_labels if label.name in labels_set]
                    check_labels = True

                if check_labels and len(shape_labels) == 0:
                    continue

                if not is_full_box:
                    # Create a denormalized copy of the shape.
                    shape = annotation.shape.denormalize_wrt_roi_shape(roi_as_box)
                else:
                    # Also create a copy of the shape, so that we can safely modify the labels
                    # without tampering with the original shape.
                    shape = copy.deepcopy(annotation.shape)

                annotations.append(Annotation(shape=shape, labels=shape_labels))
        return annotations

    def append_annotations(self, annotations: Sequence[Annotation]):
        """Adds a list of shapes to the annotation."""
        roi_as_box = ShapeFactory.shape_as_rectangle(self.roi.shape)

        validated_annotations = [
            Annotation(
                shape=annotation.shape.normalize_wrt_roi_shape(roi_as_box),
                labels=annotation.get_labels(),
            )
            for annotation in annotations
            if ShapeFactory().shape_produces_valid_crop(
                shape=annotation.shape,
                media_width=self.media.width,
                media_height=self.media.height,
            )
        ]

        n_invalid_shapes = len(annotations) - len(validated_annotations)
        if n_invalid_shapes > 0:
            logger.info(
                "%d shapes will not be added to the dataset item as they "
                "would produce invalid crops (this is expected for some tasks, "
                "such as segmentation).",
                n_invalid_shapes,
            )

        self.annotation_scene.append_annotations(validated_annotations)

    def get_roi_labels(
        self,
        labels: Optional[List[LabelEntity]] = None,
        include_empty: bool = False,
        include_ignored: bool = False,
    ) -> List[LabelEntity]:
        """Return the subset of the input labels which exist in the dataset item (wrt. ROI).

        Args:
            labels (Optional[List[LabelEntity]]): Subset of input labels to filter with; if ``None``, all the labels
                within the ROI are returned.
            include_empty (bool): if True, returns both empty and non-empty labels
            include_ignored (bool): if True, includes the labels in ignored_labels

        Return:
            List[LabelEntity]: The intersection of the input label set and those present within the ROI.
        """
        filtered_labels = set()
        for label in self.roi.get_labels(include_empty):
            if labels is None or label.get_label() in labels:
                filtered_labels.add(label.get_label())
        if not include_ignored:
            filtered_labels -= self.ignored_labels
        return sorted(list(filtered_labels), key=lambda x: x.name)

    def get_shapes_labels(
        self,
        labels: Optional[List[LabelEntity]] = None,
        include_empty: bool = False,
        include_ignored: bool = False,
    ) -> List[LabelEntity]:
        """Get the labels of the shapes present in this dataset item.

        if a label list is supplied, only labels present within that list are returned. if include_empty is True,
        present empty labels are returned as well.

        Args:
            labels (Optional[List[LabelEntity]]): if supplied only labels present in this list are returned.
                Defaults to None.
            include_empty (bool): if True, returns both empty and non-empty labels. Defaults to False.
            include_ignored (bool): if True, includes the labels in ignored_labels. Defaults to False.

        Returns:
            List[LabelEntity]: a list of labels from the shapes within the roi of this dataset item
        """
        annotations = self.get_annotations(labels=labels, include_empty=include_empty, include_ignored=include_ignored)
        scored_label_set = set(itertools.chain(*[annotation.get_labels(include_empty) for annotation in annotations]))
        label_set = {scored_label.get_label() for scored_label in scored_label_set}
        if not include_ignored:
            label_set -= self.ignored_labels
        if labels is None:
            return list(label_set)
        return [label for label in label_set if label in labels]

    def append_labels(self, labels: List[ScoredLabel]):
        """Appends labels to the DatasetItem and adds it to the the annotation label as well if it's not yet there.

        Args:
            labels (List[ScoredLabel]): list of labels to be appended.
        """
        if len(labels) == 0:
            return

        roi_annotation = None
        for annotation in self.annotation_scene.annotations:
            if annotation.shape == self.roi.shape:
                roi_annotation = annotation
                break

        if roi_annotation is None:  # no annotation found with shape
            roi_annotation = self.roi
            self.annotation_scene.append_annotation(roi_annotation)

        for label in labels:
            if label not in self.roi.get_labels(include_empty=True):
                self.roi.append_label(label)
            if label not in roi_annotation.get_labels(include_empty=True):
                roi_annotation.append_label(label)

    def __eq__(self, other):
        """Compares if two DatasetItems are equal.

        Args:
            other ("DatasetItems"): other DatasetItem to compare with.

        Returns:
            bool: True if equal, False otherwise.
        """
        if isinstance(other, DatasetItemEntity):
            return (
                self.media == other.media
                and self.annotation_scene == other.annotation_scene
                and self.roi == other.roi
                and self.subset == other.subset
                and self.ignored_labels == other.ignored_labels
            )
        return False

    def __deepcopy__(self, memo):
        """Avoids copying the lock and unintentional ID sharing among AnnotationSceneEntity instances.

        When we deepcopy this object, be sure not to deep copy the lock, as this is not possible,
        make a new lock instead. In addition, we prevent deepcopy of AnnotationSceneEntity member
        variable to avoid unintentional ID sharing among instances. Same instance reference is
        copied to the output instead.
        """
        # Call ROI getter to ensure original object has an ROI.
        _ = self.roi

        clone = copy.copy(self)

        for name, value in vars(self).items():
            if "__roi_lock" in name:
                setattr(clone, name, Lock())
            elif "__annotation_scene" in name:
                pass  # Keep the same instance
            else:
                setattr(clone, name, copy.deepcopy(value, memo))
        return clone

    def append_metadata_item(self, data: IMetadata, model: Optional[ModelEntity] = None):
        """Appends metadata produced by some model to the dataset item.

        .. rubric:: Adding visualization heatmap (ResultMediaEntity) to DatasetItemEntity

        >>> from otx.api.entities.image import Image
        >>> from otx.api.entities.result_media import ResultMediaEntity
        >>> media = Image(file_path='image.jpeg')
        >>> annotation = NullAnnotationSceneEntity()
        >>> dataset_item = DatasetItem(media=media, annotation_scene=annotation)
        >>> data = np.ones((120, 120, 3)).astype(np.uint8) * 255 # Saliency numpy
        >>> result_media = ResultMediaEntity(name="Gradcam++",
        ...                                  type="Gradcam++",
        ...                                  annotation_scene=annotation,
        ...                                  numpy=data)
        >>> dataset_item.append_metadata_item(result_media)

        .. rubric:: Representation vector for active learning

        >>> from otx.api.entities.tensor import TensorEntity
        >>> tensor = TensorEntity(name="representation_vector", numpy=data)
        >>> dataset_item.append_metadata_item(data=tensor, model=model)

        Args:
            data (IMetadata): any object of a class inherited from IMetadata. (e.g., FloatMetadata, Tensor)
            model (Optional[ModelEntity]): model that was used to generated metadata
        """
        self.__metadata.append(MetadataItemEntity(data=data, model=model))

    def get_metadata_by_name_and_model(self, name: str, model: Optional[ModelEntity]) -> Sequence[MetadataItemEntity]:
        """Returns a metadata item with `name` and generated by `model`.

        Args:
            name (str): the name of the metadata
            model (Optional[ModelEntity]): the model which was used to generate the metadata.

        Returns:
            Sequence[MetadataItemEntity]: a list of metadata items with `name` and generated by `model`.
        """
        return [meta for meta in self.get_metadata() if meta.data.name == name and meta.model == model]



"""Streamer for reading input."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
import multiprocessing
import os
import queue
import sys
from enum import Enum
from typing import Iterator, Union

import cv2
import numpy as np


class InvalidInput(Exception):
    """Exception for wrong input format."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class OpenError(Exception):
    """Exception for error opening reader."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class MediaType(Enum):
    """This Enum represents the types of input."""

    IMAGE = 1
    DIR = 2
    VIDEO = 3
    CAMERA = 4


class BaseStreamer(metaclass=abc.ABCMeta):
    """Base Streamer interface to implement Image, Video and Camera streamers."""

    @abc.abstractmethod
    def __iter__(self) -> Iterator[np.ndarray]:
        """Iterate through the streamer object that is a Python Generator object.

        Returns:
            np.ndarray: Yield the image or video frame.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_type(self) -> MediaType:
        """Get type of streamer.

        Returns:
            MediaType: type of streamer.
        """
        raise NotImplementedError


def _process_run(streamer: BaseStreamer, buffer: multiprocessing.Queue) -> None:
    """Private function that is run by the thread.

    Waits for the buffer to gain space for timeout seconds while it is full.
    If no space was available within this time the function will exit

    streamer (BaseStreamer): The streamer to retrieve frames from
    buffer (multiprocessing.Queue): The buffer to place the retrieved frames in
    """
    for frame in streamer:
        buffer.put(frame)


class ThreadedStreamer(BaseStreamer):
    """Runs a BaseStreamer on a separate thread.

    streamer (BaseStreamer): The streamer to run on a thread
    buffer_size (int): Number of frame to buffer internally. Defaults to 2.

    Example:
        >>> streamer = VideoStreamer(path="../demo.mp4")
        >>> threaded_streamer = ThreadedStreamer(streamer)
        >>> for frame in threaded_streamer:
        ...    pass
    """

    def __init__(self, streamer: BaseStreamer, buffer_size: int = 2) -> None:
        self.buffer_size = buffer_size
        self.streamer = streamer

    def __iter__(self) -> Iterator[np.ndarray]:
        """Get frames from streamer and yield them.

        Yields:
            Iterator[np.ndarray]: Yield the image or video frame.
        """
        buffer: multiprocessing.Queue = multiprocessing.Queue(maxsize=self.buffer_size)
        process = multiprocessing.Process(target=_process_run, args=(self.streamer, buffer))
        # Make thread a daemon so that it will exit when the main program exits as well
        process.daemon = True
        process.start()

        try:
            while process.is_alive() or not buffer.empty():
                try:
                    yield buffer.get(timeout=0.1)
                except queue.Empty:
                    pass
        except GeneratorExit:
            process.terminate()
        finally:
            process.join(timeout=0.1)
            # The kill() function is only available in Python 3.7.
            # Skip it if running an older Python version.
            if sys.version_info >= (3, 7) and process.exitcode is None:
                process.kill()

    def get_type(self) -> MediaType:
        """Get type of internal streamer.

        Returns:
            MediaType: type of internal streamer.
        """
        return self.streamer.get_type()


class VideoStreamer(BaseStreamer):
    """Video Streamer.

    Args:
        path: Path to the video file.

    Example:
        >>> streamer = VideoStreamer(path="../demo.mp4")
        ... for frame in streamer:
        ...    pass
    """

    def __init__(self, input_path: str, loop: bool = False) -> None:
        self.media_type = MediaType.VIDEO
        self.loop = loop
        self.cap = cv2.VideoCapture()
        status = self.cap.open(input_path)
        if not status:
            raise InvalidInput(f"Can't open the video from {input_path}")

    def __iter__(self) -> Iterator[np.ndarray]:
        """Iterates over frames of the video.

        If self.loop is set to True, the video will loop infinitely.
        """
        while True:
            status, image = self.cap.read()
            if status:
                yield cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            else:
                if self.loop:
                    self.cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
                else:
                    break

    def get_type(self) -> MediaType:
        """Returns the type of media."""
        return MediaType.VIDEO


class CameraStreamer(BaseStreamer):
    """Stream video frames from camera.

    Args:
        camera_device (int): Camera device index e.g, 0, 1

    Example:
        >>> streamer = CameraStreamer(camera_device=0)
        ... for frame in streamer:
        ...     cv2.imshow("Window", frame)
        ...     if ord("q") == cv2.waitKey(1):
        ...         break
    """

    def __init__(self, camera_device: int = 0) -> None:
        self.media_type = MediaType.CAMERA
        try:
            self.stream = cv2.VideoCapture(int(camera_device))
        except ValueError as error:
            raise InvalidInput(f"Can't find the camera {camera_device}") from error

    def __iter__(self) -> Iterator[np.ndarray]:
        """Read video and yield the frame.

        Args:
            stream: Video stream captured via OpenCV's VideoCapture

        Returns:
            Individual frame
        """
        while True:
            frame_available, frame = self.stream.read()
            if not frame_available:
                break
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            yield frame

        self.stream.release()

    def get_type(self) -> MediaType:
        """Returns the type of media."""
        return MediaType.CAMERA


class ImageStreamer(BaseStreamer):
    """Stream from image file.

    Args:
        input_path (str): Path to an image.
        loop (bool): Whether to loop through the image or not. Defaults to False.

    Example:
        >>> streamer = ImageStreamer(path="../images")
        ... for frame in streamer:
        ...     cv2.imshow("Window", frame)
        ...     cv2.waitKey(0)
    """

    def __init__(self, input_path: str, loop: bool = False) -> None:
        self.loop = loop
        self.media_type = MediaType.IMAGE
        if not os.path.isfile(input_path):
            raise InvalidInput(f"Can't find the image by {input_path}")
        self.image = cv2.imread(input_path, cv2.IMREAD_COLOR)
        if self.image is None:
            raise OpenError(f"Can't open the image from {input_path}")
        self.image = cv2.cvtColor(self.image, cv2.COLOR_BGR2RGB)

    def __iter__(self) -> Iterator[np.ndarray]:
        """If loop is True, yield the image again and again."""
        if not self.loop:
            yield self.image
        else:
            while True:
                yield self.image

    def get_type(self) -> MediaType:
        """Returns the type of the streamer."""
        return MediaType.IMAGE


class DirStreamer(BaseStreamer):
    """Stream from directory of images.

    Args:
        path: Path to directory.

    Example:
        >>> streamer = DirStreamer(path="../images")
        ... for frame in streamer:
        ...     cv2.imshow("Window", frame)
        ...     cv2.waitKey(0)
    """

    def __init__(self, input_path: str, loop: bool = False) -> None:
        self.loop = loop
        self.media_type = MediaType.DIR
        self.dir = input_path
        if not os.path.isdir(self.dir):
            raise InvalidInput(f"Can't find the dir by {input_path}")
        self.names = sorted(os.listdir(self.dir))
        if not self.names:
            raise OpenError(f"The dir {input_path} is empty")
        self.file_id = 0
        for name in self.names:
            filename = os.path.join(self.dir, name)
            image = cv2.imread(str(filename), cv2.IMREAD_COLOR)
            if image is not None:
                return
        raise OpenError(f"Can't read the first image from {input_path}")

    def __iter__(self) -> Iterator[np.ndarray]:
        """Iterates over the images in a directory.

        If self.loop is True, it reiterates again from the first image in the directory.
        """
        while self.file_id < len(self.names):
            filename = os.path.join(self.dir, self.names[self.file_id])
            image = cv2.imread(str(filename), cv2.IMREAD_COLOR)
            if self.file_id < len(self.names) - 1:
                self.file_id = self.file_id + 1
            else:
                self.file_id = self.file_id + 1 if not self.loop else 0
            if image is not None:
                yield cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    def get_type(self) -> MediaType:
        """Returns the type of the streamer."""
        return MediaType.DIR


def get_streamer(
    input_stream: Union[int, str] = 0,
    loop: bool = False,
    threaded: bool = False,
) -> BaseStreamer:
    """Get streamer object based on the file path or camera device index provided.

    Args:
        input_stream (Union[int, str]): Path to file or directory or index for camera.
        loop (bool): Enable reading the input in a loop. Defaults to False.
        threaded (bool): Run streaming on a separate thread. Threaded streaming option. Defaults to False.

    Returns:
        BaseStreamer: Streamer object.
    """
    # errors: Dict = {InvalidInput: [], OpenError: []}
    errors = []
    streamer: BaseStreamer
    for reader in (ImageStreamer, DirStreamer, VideoStreamer):
        try:
            streamer = reader(input_stream, loop)  # type: ignore
            if threaded:
                streamer = ThreadedStreamer(streamer)
            return streamer
        except (InvalidInput, OpenError) as error:
            errors.append(error)
    try:
        streamer = CameraStreamer(input_stream)  # type: ignore
        if threaded:
            streamer = ThreadedStreamer(streamer)
        return streamer
    except (InvalidInput, OpenError) as error:
        errors.append(error)

    if errors:
        raise Exception(errors)

    sys.exit(1)



"""This module contains the interface class for tasks."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc

from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters


class IExplainTask(metaclass=abc.ABCMeta):
    """A base interface for explain task."""

    @abc.abstractmethod
    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: InferenceParameters,
    ) -> DatasetEntity:
        """This is the method that is called upon explanation.

        Args:
            dataset: The input dataset to perform the explain on.
            explain_parameters: The parameters to use for the explain.

        Returns:
            The results of the explanation.
        """
        raise NotImplementedError



"""Initialization of visualizers."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .anomaly_visualizer import AnomalyVisualizer
from .visualizer import IVisualizer, Visualizer

__all__ = ["AnomalyVisualizer", "IVisualizer", "Visualizer"]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import Dict, List, Optional, Union

import openvino.runtime as ov
from mmdet.models.builder import HEADS
from mmdet.models.roi_heads.mask_heads.fcn_mask_head import FCNMaskHead

from ....mmov_model import MMOVModel


@HEADS.register_module()
class MMOVMaskHead(FCNMaskHead):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Dict[str, Union[str, List[str]]] = {},
        outputs: Dict[str, Union[str, List[str]]] = {},
        init_weight: bool = False,
        verify_shape: bool = True,
        background_index: Optional[int] = None,
        *args,
        **kwargs,
    ):
        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._inputs = deepcopy(inputs)
        self._outputs = deepcopy(outputs)
        self._init_weight = init_weight
        self._verify_sahpe = verify_shape
        self._background_index = background_index

        # dummy input
        super().__init__(*args, **kwargs)
        delattr(self, "convs")
        delattr(self, "upsample")
        delattr(self, "conv_logits")
        delattr(self, "relu")

        #  if self._background_index is not None and self._background_index < 0:
        #      self._background_index = self.num_classes + 1 - self._background_index

        self.model = MMOVModel(
            self._model_path_or_model,
            inputs=inputs,
            outputs=outputs,
            remove_normalize=False,
            merge_bn=True,
            paired_bn=True,
            init_weight=self._init_weight,
            verify_shape=self._verify_sahpe,
        )

    def init_weights(self):
        # TODO
        pass

    def forward(self, x):
        return self.model(x)



"""
Dummy Dataset for Training
"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
import shutil
from contextlib import ContextDecorator
from functools import wraps
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, List, Optional, Union

import numpy as np
from skimage.io import imsave

from .shapes import random_shapes


def generate_random_anomaly_image(
    image_width: int,
    image_height: int,
    shapes: List[str] = ["triangle", "rectangle"],
    max_shapes: int = 5,
    generate_mask: bool = False,
) -> Dict[str, np.ndarray]:
    """Generate a random image with the corresponding shape entities.

    Args:
        image_width (int): Width of the image
        image_height (int): Height of the image
        shapes (List[str]): List of shapes to draw in the image. Make sure these are different from `anomalous_shapes`
        max_shapes (int): Maximum shapes of a kind in the image. Defaults to 5.
        max_size (int): Maximum size of the test shapes. Defaults to 10.
        generate_mask (bool): Toggle between train/test split. Train images are restricted to first quadrant.
                    Also generates the mask for the image. Defaults to False.

    Returns:
        Dict[str, np.ndarray]: image if `train` is False. Else return image, mask
    """

    image: np.ndarray = np.full((image_height, image_width, 3), 255, dtype=np.uint8)
    input_region = [0, 0, image_width - 1, image_height - 1]

    for shape in shapes:
        shape_image = random_shapes(input_region, (image_height, image_width), max_shapes=max_shapes, shape=shape)
        image = np.minimum(image, shape_image)  # since white is 255

    result = {"image": image}

    if generate_mask:
        mask = np.zeros((image_height, image_width), dtype=np.uint8)
        # if color exists in any channel turn the mask bit white.
        # not sure if there is a better way to do this.
        mask[image[..., 0] < 255] = 255
        mask[image[..., 1] < 255] = 255
        mask[image[..., 2] < 255] = 255
        result["mask"] = mask

    return result


class TestDataset:
    def __init__(
        self,
        num_train: int = 1000,
        num_test: int = 100,
        img_height: int = 128,
        img_width: int = 128,
        max_size: int = 10,
        train_shapes: List[str] = ["triangle", "rectangle"],
        test_shapes: List[str] = ["hexagon", "star"],
        dataset_path: Union[str, Path] = "./datasets/MVTec",
        use_mvtec: bool = False,
    ) -> None:
        """Creates a context for Generating Dummy Dataset. Useful for wrapping test functions.
        NOTE: for MVTec dataset it does not return a category.
        It is adviced to use a default parameter in the function

        Args:
            num_train (int, optional): Number of training images to generate. Defaults to 1000.
            num_test (int, optional): Number of testing images to generate per category. Defaults to 100.
            img_height (int, optional): Height of the image. Defaults to 128.
            img_width (int, optional): Width of the image. Defaults to 128.
            max_size (Optional[int], optional): Maximum size of the test shapes. Defaults to 10.
            train_shapes (List[str], optional): List of good shapes. Defaults to ["circle", "rectangle"].
            test_shapes (List[str], optional): List of anomalous shapes. Defaults to ["triangle", "ellipse"].
            path (Union[str, Path], optional): Path to MVTec dataset. Defaults to "./datasets/MVTec".
            use_mvtec (bool, optional): Use MVTec dataset or dummy dataset. Defaults to False.

        Example:
            >>> @TestDataset
            >>> def test_some_function(path, category="leather"):
            >>>     do something
        """
        self.num_train = num_train
        self.num_test = num_test
        self.img_height = img_height
        self.img_width = img_width
        self.max_size = max_size
        self.train_shapes = train_shapes
        self.test_shapes = test_shapes
        self.dataset_path = dataset_path
        self.use_mvtec = use_mvtec

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwds):
            # If true, will use MVTech dataset for testing.
            # Useful for nightly builds
            if self.use_mvtec:
                return func(*args, dataset_path=self.dataset_path, **kwds)
            else:
                with GeneratedDummyDataset(
                    num_train=self.num_train,
                    num_test=self.num_test,
                    img_height=self.img_height,
                    img_width=self.img_width,
                    train_shapes=self.train_shapes,
                    test_shapes=self.test_shapes,
                    max_size=self.max_size,
                ) as dataset_path:
                    kwds["category"] = "shapes"
                    return func(*args, dataset_path=dataset_path, **kwds)

        return inner


class GeneratedDummyDataset(ContextDecorator):
    """Context for generating dummy shapes dataset.

    Example:
        >>> with GeneratedDummyDataset(num_train=1000,num_test=100) as dataset_path:
        >>>     some_function()

    Args:
        num_train (int, optional): Number of training images to generate. Defaults to 1000.
        num_test (int, optional): Number of testing images to generate per category. Defaults to 100.
        img_height (int, optional): Height of the image. Defaults to 128.
        img_width (int, optional): Width of the image. Defaults to 128.
        max_size (Optional[int], optional): Maximum size of the test shapes. Defaults to 10.
        train_shapes (List[str], optional): List of good shapes. Defaults to ["circle", "rectangle"].
        test_shapes (List[str], optional): List of anomalous shapes. Defaults to ["triangle", "ellipse"].
    """

    def __init__(
        self,
        num_train: int = 1000,
        num_test: int = 100,
        img_height: int = 256,
        img_width: int = 256,
        max_size: Optional[int] = 10,
        train_shapes: List[str] = ["triangle", "rectangle"],
        test_shapes: List[str] = ["star", "hexagon"],
    ) -> None:
        self.root_dir = mkdtemp()
        self.num_train = num_train
        self.num_test = num_test
        self.train_shapes = train_shapes
        self.test_shapes = test_shapes
        self.image_height = img_height
        self.image_width = img_width
        self.max_size = max_size

    def _generate_dataset(self):
        """Generates dummy dataset in a temporary directory using the same convention as MVTec"""
        # create train images
        train_path = os.path.join(self.root_dir, "shapes", "train", "good")
        os.makedirs(train_path, exist_ok=True)
        for i in range(self.num_train):
            result = generate_random_anomaly_image(
                image_width=self.image_width,
                image_height=self.image_height,
                shapes=self.train_shapes,
                generate_mask=False,
            )
            image = result["image"]
            imsave(os.path.join(train_path, f"{i:03}.png"), image, check_contrast=False)

        # create test images
        for test_category in self.test_shapes:
            test_path = os.path.join(self.root_dir, "shapes", "test", test_category)
            mask_path = os.path.join(self.root_dir, "shapes", "ground_truth", test_category)
            os.makedirs(test_path, exist_ok=True)
            os.makedirs(mask_path, exist_ok=True)
            # anomaly and masks. The idea is to superimpose anomalous shapes on top of correct ones
            for i in range(self.num_test):
                correct_shapes = generate_random_anomaly_image(
                    image_width=self.image_width,
                    image_height=self.image_height,
                    shapes=self.train_shapes,
                    generate_mask=False,
                )
                result = generate_random_anomaly_image(
                    image_width=self.image_width,
                    image_height=self.image_height,
                    shapes=[test_category],
                    generate_mask=True,
                )
                correct_shapes = correct_shapes["image"]
                image, mask = result["image"], result["mask"]
                image = np.minimum(image, correct_shapes)  # since 255 is white
                imsave(os.path.join(test_path, f"{i:03}.png"), image, check_contrast=False)
                imsave(os.path.join(mask_path, f"{i:03}_mask.png"), mask, check_contrast=False)
        # good test
        test_good = os.path.join(self.root_dir, "shapes", "test", "good")
        os.makedirs(test_good, exist_ok=True)
        for i in range(self.num_test):
            result = generate_random_anomaly_image(
                image_width=self.image_width,
                image_height=self.image_height,
                shapes=self.train_shapes,
            )
            image = result["image"]
            imsave(os.path.join(test_good, f"{i:03}.png"), image, check_contrast=False)

    def __enter__(self):
        """Creates the dataset in temp folder"""
        self._generate_dataset()
        return self.root_dir

    def __exit__(self, _exc_type, _exc_value, _exc_traceback):
        """Cleanup the directory"""
        shutil.rmtree(self.root_dir)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch
from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.rpn_head import RPNHead

from otx.mpa.utils.logger import get_logger

from ...mmov_model import MMOVModel

logger = get_logger()


@HEADS.register_module()
class MMOVRPNHead(RPNHead):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        init_weight: bool = False,
        verify_shape: bool = True,
        transpose_cls: bool = False,
        transpose_reg: bool = False,
        *args,
        **kwargs,
    ):
        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._inputs = deepcopy(inputs)
        self._outputs = deepcopy(outputs)
        self._init_weight = init_weight
        self._verify_shape = verify_shape
        self._transpose_cls = transpose_cls
        self._transpose_reg = transpose_reg

        # dummy input
        in_channels = 1
        super().__init__(in_channels=in_channels, *args, **kwargs)

    def _init_layers(self):
        self.model = MMOVModel(
            self._model_path_or_model,
            self._weight_path,
            inputs=self._inputs,
            outputs=self._outputs,
            remove_normalize=False,
            merge_bn=False,
            paired_bn=False,
            verify_shape=self._verify_shape,
            init_weight=self._init_weight,
        )

    def init_weights(self):
        # TODO
        pass

    def forward_single(self, x):
        rpn_cls_score, rpn_bbox_pred = self.model(x)

        if self._transpose_reg:
            # [B, 4 * num_anchors, H, W] -> [B, num_anchors * 4, H, W]
            shape = rpn_bbox_pred.shape
            rpn_bbox_pred = rpn_bbox_pred.reshape(shape[0], 4, -1, *shape[2:]).transpose(1, 2).reshape(shape)

        if self._transpose_cls:
            # [B, 2 * num_anchors, H, W] -> [B, num_anchors * 2, H, W]
            shape = rpn_cls_score.shape
            rpn_cls_score = rpn_cls_score.reshape(shape[0], 2, -1, *shape[2:]).transpose(1, 2).reshape(shape)

        # We set FG labels to [0, num_class-1] and BG label to
        # num_class in RPN head since mmdet v2.5, which is unified to
        # be consistent with other head since mmdet v2.0. In mmdet v2.0
        # to v2.4 we keep BG label as 0 and FG label as 1 in rpn head.
        bg = rpn_cls_score[:, 0::2]
        fg = rpn_cls_score[:, 1::2]
        rpn_cls_score = torch.flatten(torch.stack([fg, bg], dim=2), 1, 2)

        return rpn_cls_score, rpn_bbox_pred



"""This module implements helpers for drawing shapes."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-locals
import abc
from typing import (
    Callable,
    Generic,
    List,
    NewType,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import cv2
import numpy as np

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    NullAnnotationSceneEntity,
)
from otx.api.entities.coordinate import Coordinate
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.shapes.shape import ShapeEntity

CvTextSize = NewType("CvTextSize", Tuple[Tuple[int, int], int])

_Any = TypeVar("_Any")


class DrawerEntity(Generic[_Any]):
    """An interface to draw a shape of type ``T`` onto an image."""

    supported_types: Sequence[Type[ShapeEntity]] = []

    @abc.abstractmethod
    def draw(self, image: np.ndarray, entity: _Any, labels: List[ScoredLabel]) -> np.ndarray:
        """Draw an entity to a given frame.

        Args:
            image (np.ndarray): The image to draw the entity on.
            entity (T): The entity to draw.
            labels (List[ScoredLabel]): Labels of the shapes to draw

        Returns:
            np.ndarray: frame with shape drawn on it
        """
        raise NotImplementedError


class Helpers:
    """Contains variables which are used by all subclasses.

    Contains functions which help with generating coordinates, text and text scale.
    These functions are use by the DrawerEntity Classes when drawing to an image.
    """

    def __init__(self) -> None:
        # Same alpha value that the UI uses for Labels
        self.alpha_shape = 100 / 256
        self.alpha_labels = 153 / 256
        self.assumed_image_width_for_text_scale = 1280  # constant number for size of classification/counting overlay
        self.top_margin = 0.07  # part of the top screen reserved for top left classification/counting overlay
        self.content_padding = 3
        self.top_left_box_thickness = 1
        self.content_margin = 2
        self.label_offset_box_shape = 0
        self.black = (0, 0, 0)
        self.white = (255, 255, 255)
        self.yellow = (255, 255, 0)

        self.cursor_pos = Coordinate(0, 0)
        self.line_height = 0

    @staticmethod
    def draw_transparent_rectangle(
        img: np.ndarray,
        x1: int,
        y1: int,
        x2: int,
        y2: int,
        color: Tuple[int, int, int],
        alpha: float,
    ) -> np.ndarray:
        """Draw a rectangle on an image.

        Args:
            img (np.ndarray): Image
            x1 (int): Left side
            y1 (int): Top side
            x2 (int): Right side
            y2 (int): Bottom side
            color (Tuple[int, int, int]): Color
            alpha (float): Alpha value between 0 and 1
        """
        x1 = np.clip(x1, 0, img.shape[1] - 1)
        y1 = np.clip(y1, 0, img.shape[0] - 1)
        x2 = np.clip(x2 + 1, 0, img.shape[1] - 1)
        y2 = np.clip(y2 + 1, 0, img.shape[0] - 1)
        rect = img[y1:y2, x1:x2]
        rect[:] = (alpha * np.array(color))[np.newaxis, np.newaxis] + (1 - alpha) * rect
        return img

    def generate_text_scale(self, image: np.ndarray) -> float:
        """Calculates the scale of the text.

        Args:
            image (np.ndarray): Image to calculate the text scale for.

        Returns:
            scale for the text
        """
        return round(image.shape[1] / self.assumed_image_width_for_text_scale, 1)

    @staticmethod
    def generate_text_for_label(
        label: Union[LabelEntity, ScoredLabel], show_labels: bool, show_confidence: bool
    ) -> str:
        """Return a string representing a given label and its associated probability if label is a ScoredLabel.

        The exact format of the string depends on the function parameters described below.

        Args:
            label (Union[LabelEntity, ScoredLabel]): Label
            show_labels (bool): Whether to render the labels above the shape
            show_confidence (bool): Whether to render the confidence above the
                shape

        Returns:
            str: Formatted string (e.g. `"Cat 58%"`)
        """
        text = ""
        if show_labels:
            text += label.name
        if show_confidence and isinstance(label, ScoredLabel):
            if len(text) > 0:
                text += " "
            text += f"{label.probability:.0%}"
        return text

    def generate_draw_command_for_labels(
        self,
        labels: Sequence[Union[LabelEntity, ScoredLabel]],
        image: np.ndarray,
        show_labels: bool,
        show_confidence: bool,
    ) -> Tuple[Callable[[np.ndarray], np.ndarray], int, int]:
        """Generate draw function and content width and height for labels.

        Generates a function which can be called to draw a list of labels onto an image relatively to the
        cursor position.
        The width and height of the content is also returned and can be determined to compute
        the best position for content before actually drawing it.

        Args:
            labels (Sequence[Union[LabelEntity, ScoredLabel]]): List of labels
            image (np.ndarray): Image (used to compute font size)
            show_labels (bool): Whether to show the label name
            show_confidence (bool): Whether to show the confidence probability

        Returns:
            A tuple containing the drawing function, the content width,
            and the content height
        """
        draw_commands = []
        content_width = 0
        content_height = 0

        # Loop through the list of labels and create a function which can be used to draw the label.
        for label in labels:
            text = self.generate_text_for_label(label, show_labels, show_confidence)
            text_scale = self.generate_text_scale(image)
            thickness = int(text_scale / 2)
            color = label.color.bgr_tuple

            item_command, item_width, item_height = self.generate_draw_command_for_text(
                text, text_scale, thickness, color
            )

            draw_commands.append(item_command)

            content_width += item_width
            content_height = max(content_height, item_height)

        def draw_command(img: np.ndarray) -> np.ndarray:
            for command in draw_commands:
                img = command(img)
            return img

        return draw_command, content_width, content_height

    def generate_draw_command_for_text(
        self, text: str, text_scale: float, thickness: int, color: Tuple[int, int, int]
    ) -> Tuple[Callable[[np.ndarray], np.ndarray], int, int]:
        """Generate function to draw text on image relative to cursor position.

        Generate a function which can be called to draw the given text onto an image
        relatively to the cursor position.

        The width and height of the content is also returned and can be determined to compute
        the best position for content before actually drawing it.

        Args:
            text (str): Text to draw
            text_scale (float): Font size
            thickness (int): Thickness of the text
            color (Tuple[int, int, int]): Color of the text

        Returns:
            A tuple containing the drawing function, the content width,
            and the content height
        """

        padding = self.content_padding
        margin = self.content_margin

        label_size = cv2.getTextSize(text, cv2.FONT_HERSHEY_SIMPLEX, fontScale=text_scale, thickness=thickness)

        baseline = label_size[1]
        text_width = label_size[0][0]
        text_height = label_size[0][1]

        width = text_width + 2 * padding
        height = text_height + baseline + 2 * padding
        content_width = width + margin

        if (color[0] + color[1] + color[2]) / 3 > 200:
            text_color = self.black
        else:
            text_color = self.white

        def draw_command(img: np.ndarray) -> np.ndarray:
            cursor_pos = Coordinate(int(self.cursor_pos.x), int(self.cursor_pos.y))
            self.draw_transparent_rectangle(
                img,
                int(cursor_pos.x),
                int(cursor_pos.y),
                int(cursor_pos.x + width),
                int(cursor_pos.y + height),
                color,
                self.alpha_labels,
            )

            img = cv2.putText(
                img=img,
                text=text,
                org=(
                    cursor_pos.x + padding,
                    cursor_pos.y + height - padding - baseline,
                ),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=text_scale,
                color=text_color,
                thickness=thickness,
                lineType=cv2.LINE_AA,
            )

            self.cursor_pos.x += content_width
            self.line_height = height

            return img

        return draw_command, content_width, height

    @staticmethod
    def draw_flagpole(
        image: np.ndarray,
        flagpole_start_point: Coordinate,
        flagpole_end_point: Coordinate,
    ):
        """Draw a small flagpole between two points.

        Args:
            image: Image
            flagpole_start_point: Start of the flagpole
            flagpole_end_point: End of the flagpole

        Returns:
            Image
        """
        return cv2.line(
            image,
            flagpole_start_point.as_int_tuple(),
            flagpole_end_point.as_int_tuple(),
            color=[0, 0, 0],
            thickness=2,
        )

    def newline(self):
        """Move the cursor to the next line."""
        self.cursor_pos.x = 0
        self.cursor_pos.y += self.line_height + self.content_margin

    def set_cursor_pos(self, cursor_pos: Optional[Coordinate] = None):
        """Move the cursor to a new position.

        Args:
            cursor_pos (Optional[Coordinate]): New position of the cursor; (0,0) if not specified.
        """
        if cursor_pos is None:
            cursor_pos = Coordinate(0, 0)

        self.cursor_pos = cursor_pos


class ShapeDrawer(DrawerEntity[AnnotationSceneEntity]):
    """ShapeDrawer to draw any shape on a numpy array. Will overlay the shapes in the same way that the UI does.

    Args:
        show_count: Whether or not to render the amount of objects on
            screen in the top left.
        is_one_label: Whether there is only one label present in the
            project.
    """

    # TODO Connect show_count,is_is_one_label to the UI for toggling.
    def __init__(self, show_count, is_one_label):
        super().__init__()
        self.show_labels = True
        self.show_confidence = True
        self.show_count = show_count
        self.is_one_label = is_one_label

        if self.is_one_label and not self.show_count:
            self.show_labels = False

        self.shape_drawers = [
            self.RectangleDrawer(self.show_labels, self.show_confidence),
            self.PolygonDrawer(self.show_labels, self.show_confidence),
            self.EllipseDrawer(self.show_labels, self.show_confidence),
        ]

        # Always show global labels, especially if shape labels are disabled (because of is_one_label).
        self.top_left_drawer = self.TopLeftDrawer(True, self.show_confidence, self.is_one_label)

    def draw(
        self,
        image: np.ndarray,
        entity: AnnotationSceneEntity,
        labels: List[ScoredLabel],
    ) -> np.ndarray:
        """Use a compatible drawer to draw all shapes of an annotation to the corresponding image.

        Also render a label in the top left if we need to.

        Args:
            image: Numpy image, one frame of a video on which to draw
                something
            entity: AnnotationSceneEntity entity corresponding to this
                particular frame of the video
            labels: Can be passed as an empty list since they are
                already present in annotation_scene

        Returns:
            Modified image.
        """

        num_annotations = 0

        self.top_left_drawer.set_cursor_pos()

        if not isinstance(entity, NullAnnotationSceneEntity):
            for annotation in entity.annotations:
                if (
                    isinstance(annotation.shape, Rectangle)
                    and annotation.shape.x1 == 0
                    and annotation.shape.y1 == 0
                    and annotation.shape.x2 == 1
                    and annotation.shape.y2 == 1
                ):
                    # If is_one_label is activated, don't draw the labels here
                    # because we will draw them again outside the loop.
                    if not self.is_one_label:
                        image = self.top_left_drawer.draw(image, annotation, labels=[])
                else:
                    num_annotations += 1
                    for drawer in self.shape_drawers:
                        if type(annotation.shape) in drawer.supported_types and len(annotation.get_labels()) > 0:
                            image = drawer.draw(image, annotation.shape, labels=annotation.get_labels())
            if self.is_one_label:
                image = self.top_left_drawer.draw_labels(image, entity.get_labels())
            if self.show_count:
                image = self.top_left_drawer.draw_annotation_count(image, num_annotations)
        return image

    class TopLeftDrawer(Helpers, DrawerEntity[Annotation]):
        """Draws labels in an image's top left corner."""

        def __init__(self, show_labels, show_confidence, is_one_label):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence
            self.is_one_label = is_one_label

        def draw(self, image: np.ndarray, entity: Annotation, labels: List[ScoredLabel]) -> np.ndarray:
            """Draw the labels of a shape in the image top left corner.

            Args:
                image (np.ndarray): Image
                entity (Annotation): Annotation
                labels (List[ScoredLabels]): (Unused) labels to be drawn on the image

            Returns:
                np.ndarray: Image with label on top.
            """
            return self.draw_labels(image, entity.get_labels())

        def draw_labels(self, image: np.ndarray, labels: Sequence[Union[LabelEntity, ScoredLabel]]) -> np.ndarray:
            """Draw the labels in the image top left corner.

            Args:
                image (np.ndarray): Image
                labels (Sequence[Union[LabelEntity, ScoredLabel]]): Sequence of labels

            Returns:
                np.ndarray: Image with label on top.
            """
            show_confidence = self.show_confidence if not self.is_one_label else False

            draw_command, _, _ = self.generate_draw_command_for_labels(labels, image, self.show_labels, show_confidence)

            image = draw_command(image)

            if len(labels) > 0:
                self.newline()

            return image

        def draw_annotation_count(self, image: np.ndarray, num_annotations: int) -> np.ndarray:
            """Draw the number of annotations to the top left corner of the image.

            Args:
                image (np.ndarray): Image
                num_annotations (int): Number of annotations

            Returns:
                np.ndarray: Image with annotation count on top.
            """
            text = f"Count: {num_annotations}"
            color = self.yellow

            text_scale = self.generate_text_scale(image)
            draw_command, _, _ = self.generate_draw_command_for_text(
                text, text_scale, self.top_left_box_thickness, color
            )
            image = draw_command(image)

            self.newline()

            return image

    class RectangleDrawer(Helpers, DrawerEntity[Rectangle]):
        """Draws rectangles."""

        supported_types = [Rectangle]

        def __init__(self, show_labels, show_confidence):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence

        def draw(self, image: np.ndarray, entity: Rectangle, labels: List[ScoredLabel]) -> np.ndarray:
            """Draws a rectangle on the image along with labels.

            Args:
                image (np.ndarray): Image to draw on.
                entity (Rectangle): Rectangle to draw.
                labels (List[ScoredLabel]): List of labels.

            Returns:
                np.ndarray: Image with rectangle drawn on it.
            """
            base_color = labels[0].color.bgr_tuple

            # Draw the rectangle on the image
            x1, y1 = int(entity.x1 * image.shape[1]), int(entity.y1 * image.shape[0])
            x2, y2 = int(entity.x2 * image.shape[1]), int(entity.y2 * image.shape[0])
            image = self.draw_transparent_rectangle(image, x1, y1, x2, y2, base_color, self.alpha_shape)
            image = cv2.rectangle(img=image, pt1=(x1, y1), pt2=(x2, y2), color=base_color, thickness=2)

            (
                draw_command,
                content_width,
                content_height,
            ) = self.generate_draw_command_for_labels(labels, image, self.show_labels, self.show_confidence)

            # Generate a command to draw the list of labels
            # and compute the actual size of the list of labels.
            y_coord = y1 - self.label_offset_box_shape - content_height
            x_coord = x1

            # put label inside if it is out of bounds at the top of the shape, and shift label to left if needed
            if y_coord < self.top_margin * image.shape[0]:
                y_coord = y1 + self.label_offset_box_shape
            if x_coord + content_width > image.shape[1]:
                x_coord = x2 - content_width

            # Draw the list of labels.
            self.set_cursor_pos(Coordinate(x_coord, y_coord))
            image = draw_command(image)
            return image

    class EllipseDrawer(Helpers, DrawerEntity[Ellipse]):
        """Draws ellipses."""

        supported_types = [Ellipse]

        def __init__(self, show_labels, show_confidence):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence

        def draw(self, image: np.ndarray, entity: Ellipse, labels: List[ScoredLabel]) -> np.ndarray:
            """Draw the ellipse on the image.

            Args:
                image (np.ndarray): Image to draw on.
                entity (Ellipse): Ellipse to draw.
                labels (List[ScoredLabel]): Labels to draw.

            Returns:
                np.ndarray: Image with the ellipse drawn on it.
            """
            base_color = labels[0].color.bgr_tuple
            if entity.width > entity.height:
                axes = (
                    int(entity.major_axis * image.shape[1]),
                    int(entity.minor_axis * image.shape[0]),
                )
            else:
                axes = (
                    int(entity.major_axis * image.shape[0]),
                    int(entity.minor_axis * image.shape[1]),
                )
            center = (
                int(entity.x_center * image.shape[1]),
                int(entity.y_center * image.shape[0]),
            )
            # Draw the shape on the image
            alpha = self.alpha_shape
            overlay = cv2.ellipse(
                img=image.copy(),
                center=center,
                axes=axes,
                angle=0,
                startAngle=0,
                endAngle=360,
                color=base_color,
                thickness=cv2.FILLED,
            )
            result_without_border = cv2.addWeighted(overlay, alpha, image, 1 - alpha, 0)
            result_with_border = cv2.ellipse(
                img=result_without_border,
                center=center,
                axes=axes,
                angle=0,
                startAngle=0,
                endAngle=360,
                color=base_color,
                lineType=cv2.LINE_AA,
            )

            # Generate a command to draw the list of labels
            # and compute the actual size of the list of labels.
            (
                draw_command,
                content_width,
                content_height,
            ) = self.generate_draw_command_for_labels(labels, image, self.show_labels, self.show_confidence)

            # get top left corner of imaginary bbox around circle
            offset = self.label_offset_box_shape
            x_coord = entity.x1 * image.shape[1]
            y_coord = entity.y1 * image.shape[0] - offset - content_height

            flagpole_end_point = Coordinate(int(x_coord + 1), int(entity.y_center * image.shape[0]))

            # put label at bottom if it is out of bounds at the top of the shape, and shift label to left if needed
            if y_coord < self.top_margin * image.shape[0]:
                y_coord = (entity.y1 * image.shape[0]) + (entity.y2 * image.shape[0]) + offset
                flagpole_start_point = Coordinate(x_coord + 1, y_coord)
            else:
                flagpole_start_point = Coordinate(x_coord + 1, y_coord + content_height)

            if x_coord + content_width > result_with_border.shape[1]:
                # The list of labels is too close to the right side of the image.
                # Move it slightly to the left.
                x_coord = result_with_border.shape[1] - content_width

            # Draw the list of labels and a small flagpole.
            self.set_cursor_pos(Coordinate(x_coord, y_coord))
            image = draw_command(result_with_border)
            image = self.draw_flagpole(image, flagpole_start_point, flagpole_end_point)

            return image

    class PolygonDrawer(Helpers, DrawerEntity[Polygon]):
        """Draws polygons."""

        supported_types = [Polygon]

        def __init__(self, show_labels, show_confidence):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence

        def draw(self, image: np.ndarray, entity: Polygon, labels: List[ScoredLabel]) -> np.ndarray:
            """Draw polygon and labels on image.

            Args:
                image (np.ndarray): Image to draw on.
                entity (Polygon): Polygon to draw.
                labels (List[ScoredLabel]): List of labels to draw.

            Returns:
                np.ndarray: Image with polygon drawn on it.
            """
            base_color = labels[0].color.bgr_tuple

            # Draw the shape on the image
            alpha = self.alpha_shape
            contours = np.array(
                [[point.x * image.shape[1], point.y * image.shape[0]] for point in entity.points],
                dtype=np.int32,
            )
            overlay = cv2.drawContours(
                image=image.copy(),
                contours=[contours],
                contourIdx=-1,
                color=base_color,
                thickness=cv2.FILLED,
            )
            result_without_border = cv2.addWeighted(overlay, alpha, image, 1 - alpha, 0)
            result_with_border = cv2.drawContours(
                image=result_without_border,
                contours=[contours],
                contourIdx=-1,
                color=base_color,
                thickness=2,
                lineType=cv2.LINE_AA,
            )

            # Generate a command to draw the list of labels
            # and compute the actual size of the list of labels.
            (
                draw_command,
                content_width,
                content_height,
            ) = self.generate_draw_command_for_labels(labels, image, self.show_labels, self.show_confidence)

            # get top left corner of imaginary bbox around polygon
            x_coord = min(point[0] for point in contours)
            y_coord = min(point[1] for point in contours) - self.label_offset_box_shape - content_height

            # end point = Y in polygon where X is lowest, x offset to make line flush with text rectangle
            _, idx = min((val, idx) for (idx, val) in enumerate([point[0] for point in contours]))
            flagpole_end_point = Coordinate(x_coord + 1, [point[1] for point in contours][idx])

            if y_coord < self.top_margin * image.shape[0]:
                # The polygon is too close to the top of the image.
                # Draw the labels underneath the polygon instead.
                y_coord = max(point[1] for point in contours) + self.label_offset_box_shape
                flagpole_start_point = Coordinate(x_coord + 1, y_coord)
            else:
                flagpole_start_point = Coordinate(x_coord + 1, y_coord + content_height)

            if x_coord + content_width > result_with_border.shape[1]:
                # The list of labels is too close to the right side of the image.
                # Move it slightly to the left.
                x_coord = result_with_border.shape[1] - content_width

            # Draw the list of labels and a small flagpole.
            self.set_cursor_pos(Coordinate(x_coord, y_coord))
            image = draw_command(result_with_border)
            image = self.draw_flagpole(image, flagpole_start_point, flagpole_end_point)

            return image



"""Collections of Utils for common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import importlib
import inspect
from collections import defaultdict
from typing import Callable, Literal, Optional, Tuple

import yaml

from otx.api.utils.argument_checks import YamlFilePathCheck, check_input_parameters_type


class UncopiableDefaultDict(defaultdict):
    """Defauldict type object to avoid deepcopy."""

    def __deepcopy__(self, memo):
        """Deepcopy."""
        return self


@check_input_parameters_type({"path": YamlFilePathCheck})
def load_template(path):
    """Loading model template function."""
    with open(path, encoding="UTF-8") as f:
        template = yaml.safe_load(f)
    return template


@check_input_parameters_type()
def get_task_class(path: str):
    """Return Task classes."""
    module_name, class_name = path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


@check_input_parameters_type()
def get_arg_spec(  # noqa: C901  # pylint: disable=too-many-branches
    fn: Callable,  # pylint: disable=invalid-name
    depth: Optional[int] = None,
) -> Tuple[str, ...]:
    """Get argument spec of function."""

    args = set()

    cls_obj = None
    if inspect.ismethod(fn):
        fn_name = fn.__name__
        cls_obj = fn.__self__
        if not inspect.isclass(cls_obj):
            cls_obj = cls_obj.__class__
    else:
        fn_name = fn.__name__
        names = fn.__qualname__.split(".")
        if len(names) > 1 and names[-1] == fn_name:
            cls_obj = globals()[".".join(names[:-1])]

    if cls_obj:
        for obj in cls_obj.mro():  # type: ignore
            fn_obj = cls_obj.__dict__.get(fn_name, None)
            if fn_obj is not None:
                if isinstance(fn_obj, staticmethod):
                    cls_obj = None
                    break

    if cls_obj is None:
        # function, staticmethod
        spec = inspect.getfullargspec(fn)
        args.update(spec.args)
    else:
        # method, classmethod
        for i, obj in enumerate(cls_obj.mro()):  # type: ignore
            if depth is not None and i == depth:
                break
            method = getattr(obj, fn_name, None)
            if method is None:
                break
            spec = inspect.getfullargspec(method)
            args.update(spec.args[1:])
            if spec.varkw is None and spec.varargs is None:
                break
    return tuple(args)


def left_vlaue_is_better(val1, val2, mode: Literal["max", "min"]) -> bool:
    """Check left value is better than right value.

    Whether check it's greather or lesser is changed depending on 'model'.

    Args:
        val1 : value to check that it's bigger than other value.
        val2 : value to check that it's bigger than other value.
        mode (Literal['max', 'min']): value to decide whether better means greater or lesser.

    Returns:
        bool: whether val1 is better than val2.
    """
    check_mode_input(mode)
    if mode == "max":
        return val1 > val2
    return val1 < val2


def check_positive(value, variable_name: Optional[str] = None, error_message: Optional[str] = None):
    """Validate that value is positivle.

    Args:
        value (Any): value to validate.
        variable_name (Optional[str], optional): name of value. It's used for error message. Defaults to None.
        error_message (Optional[str], optional): Error message to use when type is different. Defaults to None.

    Raises:
        ValueError: If value isn't positive, the error is raised.
    """
    if value <= 0:
        if error_message is not None:
            message = error_message
        elif variable_name:
            message = f"{variable_name} should be positive.\n" f"your value : {value}"
        else:
            raise ValueError
        raise ValueError(message)


def check_not_negative(value, variable_name: Optional[str] = None, error_message: Optional[str] = None):
    """Validate that value isn't negative.

    Args:
        value (Any): value to validate.
        variable_name (Optional[str], optional): name of value. It's used for error message. Defaults to None.
        error_message (Optional[str], optional): Error message to use when type is different. Defaults to None.

    Raises:
        ValueError: If value is negative, the error is raised.
    """
    if value < 0:
        if error_message is not None:
            message = error_message
        elif variable_name:
            message = f"{variable_name} should be positive.\n" f"your value : {value}"
        else:
            raise ValueError
        raise ValueError(message)


def check_mode_input(mode: str):
    """Validate that mode is 'max' or 'min'.

    Args:
        mode (str): string to validate.

    Raises:
        ValueError: If 'mode' is not both 'max' and 'min', the error is raised.
    """
    if mode not in ["max", "min"]:
        raise ValueError("mode should be max or min.\n" f"Your value : {mode}")



"""This module implements the TrainingParameters entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass
from typing import Any, Callable, Optional, Protocol


class UpdateProgressCallback(Protocol):
    """UpdateProgressCallback protocol.

    Used as a replacement of Callable[] type since Callable doesn’t handle default parameters like
    `score: Optional[float] = None`
    """

    def __call__(self, progress: int, score: Optional[float] = None):
        """Callback to provide updates about the progress of a task.

        It is recommended to call this function at least once per epoch.
        However, the exact frequency is left to the task implementer.

        An optional `score` can also be passed. If specified, this score can be used by HPO
        to monitor the improvement of the task.

        Args:
            progress: Progress as a percentage
            score: Optional validation score
        """


# pylint: disable=unused-argument
def default_progress_callback(progress: int, score: Optional[float] = None):
    """Default progress callback. It is a placeholder (does nothing) and is used in empty TrainParameters."""


def default_save_model_callback():
    """Default save model callback. It is a placeholder (does nothing) and is used in empty TrainParameters."""


@dataclass
class TrainParameters:
    """Train parameters.

    Attributes:
        resume: Set to ``True`` if training must be resume with the
            optimizer state; set to ``False`` to discard the optimizer
            state and start with fresh optimizer
        update_progress: Callback which can be used to provide updates
            about the progress of a task.
        save_model: Callback to notify that the model weights have been
            changed. This callback can be used by the task when
            temporary weights should be saved (for instance, at the end
            of an epoch). If this callback has been used to save
            temporary weights, those weights will be used to resume
            training if for some reason training was suspended.
    """

    resume: bool = False
    update_progress: Callable[[int, Optional[float]], Any] = default_progress_callback
    save_model: Callable[[], None] = default_save_model_callback



"""Sample Code of otx training for detection."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import sys
from random import randint

import numpy as np
from mmcv.utils import get_logger

from otx.algorithms.common.utils import get_task_class
from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
    NullAnnotationSceneEntity,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType

logger = get_logger(name="mmdet")


def parse_args():
    """Parse function for getting model template & check export."""
    parser = argparse.ArgumentParser(description="Sample showcasing the new API")
    parser.add_argument("template_file_path", help="path to template file")
    parser.add_argument("--export", action="store_true")
    return parser.parse_args()


colors = dict(red=(255, 0, 0), green=(0, 255, 0))


def load_test_dataset():
    """Load Sample dataset for detection."""

    def gen_image(resolution, x1, y1, x2, y2, color):
        width, height = resolution
        image = np.full([height, width, 3], fill_value=255, dtype=np.uint8)
        image[int(y1 * height) : int(y2 * height), int(x1 * width) : int(x2 * width), :] = np.full(
            [int(height * y2) - int(height * y1), int(width * x2) - int(width * x1), 3],
            fill_value=colors[color],
            dtype=np.uint8,
        )
        return (image, Rectangle(x1=x1, y1=y1, x2=x2, y2=y2))

    labels = [
        LabelEntity(name="red", domain=Domain.DETECTION, id=0),  # OLD class
        LabelEntity(name="green", domain=Domain.DETECTION, id=1),
    ]

    def get_image(subset, label_id):
        def get_randcoord():
            x1 = randint(0, 9)
            y1 = randint(0, 9)
            x2 = min(x1 + 2, 10)
            y2 = min(y1 + 2, 10)
            coord = (x1 / 10, y1 / 10, x2 / 10, y2 / 10)
            return coord

        coord = get_randcoord()
        image, bbox = gen_image((640, 480), *coord, labels[label_id].name)
        if subset != Subset.UNLABELED:
            return DatasetItemEntity(
                media=Image(data=image),
                annotation_scene=AnnotationSceneEntity(
                    annotations=[Annotation(bbox, labels=[ScoredLabel(label=labels[label_id])])],
                    kind=AnnotationSceneKind.ANNOTATION,
                ),
                subset=subset,
            )
        return DatasetItemEntity(
            media=Image(data=image),
            annotation_scene=NullAnnotationSceneEntity(),
            subset=subset,
        )

    train = [get_image(Subset.TRAINING, 0) for i in range(10)]
    train += [get_image(Subset.TRAINING, 1) for i in range(10)]
    val = [get_image(Subset.VALIDATION, 0) for i in range(2)]
    val += [get_image(Subset.VALIDATION, 1) for i in range(2)]
    val += [get_image(Subset.TESTING, 0) for i in range(2)]
    val += [get_image(Subset.TESTING, 1) for i in range(2)]
    unlabeled = [get_image(Subset.UNLABELED, 0) for i in range(100)]
    unlabeled += [get_image(Subset.UNLABELED, 1) for i in range(100)]

    return DatasetEntity(train + val + unlabeled), labels


# pylint: disable=too-many-locals, too-many-statements
def main(args):
    """Main function of Detection Sample."""
    logger.info("[Semi-SL] Train  model with unlabeled dataset")
    dataset, labels_list = load_test_dataset()
    labels_schema = LabelSchemaEntity.from_labels(labels_list)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5
    params.learning_parameters.learning_rate = 0.01
    params.learning_parameters.learning_rate_warmup_iters = 1
    params.learning_parameters.batch_size = 4

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=None,
        hyper_parameters=params,
        label_schema=labels_schema,
        model_template=model_template,
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    output_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, output_model)

    logger.info("Get predictions on the validation set")
    validation_dataset = dataset.get_subset(Subset.VALIDATION)
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(),
        InferenceParameters(is_evaluation=False),
    )
    resultset = ResultSetEntity(
        model=output_model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    logger.info("Estimate quality on validation set")
    task.evaluate(resultset)
    logger.info(str(resultset.performance))

    if args.export:
        logger.info("Export model")
        exported_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        task.export(ExportType.OPENVINO, exported_model)

        logger.info("Create OpenVINO Task")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        openvino_task_cls = get_task_class(openvino_task_impl_path)
        openvino_task = openvino_task_cls(environment)

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )

        resultset = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Estimate quality on validation set")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Run POT optimization")
        optimized_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        openvino_task.optimize(
            OptimizationType.POT,
            dataset.get_subset(Subset.TRAINING),
            optimized_model,
            OptimizationParameters(),
        )

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )
        resultset = ResultSetEntity(
            model=optimized_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Performance of optimized model:")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Running the NNCF optimization")
        environment.model = output_model
        nncf_task_impl_path = model_template.entrypoints.nncf
        nncf_task_cls = get_task_class(nncf_task_impl_path)
        nncf_task = nncf_task_cls(environment)

        optimized_model = ModelEntity(
            dataset,
            configuration=environment.get_model_configuration(),
        )
        nncf_task.optimize(OptimizationType.NNCF, dataset, optimized_model)

        logger.info("Inferring the optimised model on the validation set.")
        predicted_validation_dataset = nncf_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )
        resultset = ResultSetEntity(
            model=optimized_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )

        logger.info("Evaluating the optimized model on the validation set.")
        nncf_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Exporting the model.")
        exported_model = ModelEntity(
            train_dataset=dataset,
            configuration=environment.get_model_configuration(),
        )
        nncf_task.export(ExportType.OPENVINO, exported_model)
        environment.model = exported_model

        logger.info("Creating the OpenVINO Task.")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        nncf_openvino_task_cls = get_task_class(openvino_task_impl_path)
        nncf_openvino_task = nncf_openvino_task_cls(environment)

        logger.info("Inferring the exported model on the validation set.")
        predicted_validation_dataset = nncf_openvino_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )

        logger.info("Evaluating the exported model on the validation set.")
        resultset = ResultSetEntity(
            model=exported_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        nncf_openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))


if __name__ == "__main__":
    sys.exit(main(parse_args()) or 0)



"""Register custom modules for mmcls models."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.algorithms.common.adapters.nncf.utils import is_nncf_enabled

if is_nncf_enabled():
    from nncf.torch import register_module
    from timm.models.layers.conv2d_same import Conv2dSame

    # Register custom modules.
    # Users of nncf should manually check every custom
    # layer with weights which should be compressed and
    # in case such layers are not wrapping by nncf,
    # wrap such custom module by yourself.
    register_module(ignored_algorithms=[])(Conv2dSame)
